import os
import shutil
import tempfile
from contextlib import contextmanager
from pathlib import Path

import h5py
import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_almost_equal, assert_array_equal
from pandas.api.types import CategoricalDtype
from pandas.testing import assert_frame_equal, assert_series_equal

import voxcell.cell_collection as test_module
from voxcell import VoxcellError

DATA_PATH = Path(__file__).parent / 'data'
SONATA_DATA_PATH = DATA_PATH / 'sonata'


def is_cat(series):
    return isinstance(series.dtype, pd.CategoricalDtype)


def euler_to_matrix(bank, attitude, heading):
    """Build 3x3 rotation matrices from arrays of euler angles.

    Based on algorigthm described here:
    http://www.euclideanspace.com/maths/geometry/rotations/conversions/eulerToMatrix/index.htm

    Args:
        bank: rotation around X
        attitude: rotation around Z
        heading: rotation around Y
    """

    sa, ca = np.sin(attitude), np.cos(attitude)
    sb, cb = np.sin(bank), np.cos(bank)
    sh, ch = np.sin(heading), np.cos(heading)

    m = np.vstack([ch * ca, -ch * sa * cb + sh * sb, ch * sa * sb + sh * cb,
                   sa, ca * cb, -ca * sb,
                   -sh * ca, sh * sa * cb + ch * sb, -sh * sa * sb + ch * cb]).transpose()

    return m.reshape(m.shape[:-1] + (3, 3))


def test_euler_to_matrix():  # testing the test
    n = 2

    assert_array_equal(
        euler_to_matrix([0] * n, [0] * n, [0] * n),
        np.array([np.diag([1, 1, 1])] * n))

    assert_almost_equal(
        euler_to_matrix([np.deg2rad(90)] * n, [0] * n, [0] * n),
        np.array([np.array([[1, 0, 0],
                            [0, 0, -1],
                            [0, 1, 0]])] * n))

    assert_almost_equal(
        euler_to_matrix([0] * n, [np.deg2rad(90)] * n, [0] * n),
        np.array([np.array([[0, -1, 0],
                            [1, 0, 0],
                            [0, 0, 1]])] * n))

    assert_almost_equal(
        euler_to_matrix([0] * n, [0] * n, [np.deg2rad(90)] * n),
        np.array([np.array([[0, 0, 1],
                            [0, 1, 0],
                            [-1, 0, 0]])] * n))


def random_orientations(n):
    return euler_to_matrix(np.random.random(n) * np.pi * 2,
                           np.random.random(n) * np.pi * 2,
                           np.random.random(n) * np.pi * 2)


def random_positions(n):
    return np.random.random((n, 3))


@contextmanager
def tempcwd():
    """Tempdir contextmanager."""
    cwd = os.getcwd()
    dirname = tempfile.mkdtemp(prefix='bbtests_')
    os.chdir(dirname)
    try:
        yield dirname
    finally:
        os.chdir(cwd)
        shutil.rmtree(dirname)


def assert_equal_cells(c0, c1):
    if c0.positions is None:
        assert c1.positions is None
    else:
        assert_almost_equal(c0.positions, c1.positions)
    if c0.orientations is None:
        assert c1.orientations is None
    else:
        assert_almost_equal(c0.orientations, c1.orientations)
    sorted_c0 = c0.properties.sort_index(axis=1)
    sorted_c1 = c1.properties.sort_index(axis=1)
    assert_array_equal(sorted_c0.columns, sorted_c1.columns)
    for column in sorted_c0:
        s_c0, s_c1 = sorted_c0[column], sorted_c1[column]
        if is_cat(s_c0) or is_cat(s_c1):
            assert_array_equal(s_c0.to_numpy(), s_c1.to_numpy())
            assert_array_equal(s_c0.index.to_numpy(), s_c1.index.to_numpy())
        else:
            assert_series_equal(s_c0, s_c1, check_names=True)


def check_roundtrip(original):
    with tempcwd():
        original.save('cells.mvd3')
        restored = test_module.CellCollection.load('cells.mvd3')
        assert_equal_cells(original, restored)
        restored_mvd3 = test_module.CellCollection.load_mvd3('cells.mvd3')
        assert_equal_cells(restored, restored_mvd3)
        original.save('cells.h5')
        restored = test_module.CellCollection.load('cells.h5')
        assert_equal_cells(original, restored)
        restored_sonata = test_module.CellCollection.load_sonata('cells.h5')
        assert_equal_cells(restored, restored_sonata)
        return restored


def test_is_string_enum():
    str_series = pd.Series(list('babc')).astype(CategoricalDtype(list('abcd')))
    assert test_module._is_string_enum(str_series)
    int_series = pd.Series([1, 2, 3, 1], dtype="category")
    assert not test_module._is_string_enum(int_series)
    float_series = pd.Series(pd.Categorical.from_codes(codes=[0, 0, 1, 0], categories=[1.5, -.4]))
    assert not test_module._is_string_enum(float_series)


def test_load_mvd2():
    cells_mvd3 = test_module.CellCollection.load_mvd3(DATA_PATH / 'mvd2_mvd3/circuit.mvd3')
    cells_mvd2 = test_module.CellCollection.load_mvd2(DATA_PATH / 'mvd2_mvd3/circuit.mvd2')
    assert_equal_cells(cells_mvd2, cells_mvd3)


def test_roundtrip_mvd():
    cells = test_module.CellCollection.load_mvd2(DATA_PATH / 'mvd2_mvd3/circuit.mvd2')
    check_roundtrip(cells)


def test_roundtrip_empty():
    cells = test_module.CellCollection()
    check_roundtrip(cells)


def test_roundtrip_none():
    cells = test_module.CellCollection()
    cells.properties['y-factor'] = [0.25, np.nan, 0.75]
    with pytest.raises(VoxcellError):
        check_roundtrip(cells)
    cells.properties['y-factor'] = [None, 0.1, 0.75]
    with pytest.raises(VoxcellError):
        check_roundtrip(cells)


def test_roundtrip_properties_numeric_single():
    cells = test_module.CellCollection()
    cells.properties['y-factor'] = [0.25, 0.5, 0.75]
    check_roundtrip(cells)


def test_roundtrip_properties_numeric_multiple():
    cells = test_module.CellCollection()
    cells.properties['y-factor'] = [0.25, 0.5, 0.75, 0]
    cells.properties['z-factor'] = [0, 0.75, 0.5, 0.25]
    check_roundtrip(cells)


def test_roundtrip_properties_text_single():
    cells = test_module.CellCollection()
    cells.properties['y-type'] = ['pretty', 'ugly', 'pretty']
    restored = check_roundtrip(cells)
    restored.properties['y-type'].to_frame()


def test_roundtrip_properties_text_multiple():
    cells = test_module.CellCollection()
    cells.properties['y-type'] = ['ugly', 'pretty', 'ugly', 'ugly']
    cells.properties['z-type'] = ['red', 'blue', 'green', 'alpha']
    restored = check_roundtrip(cells)
    restored.properties['y-type'].to_frame()
    restored.properties['z-type'].to_frame()


def test_roundtrip_properties_text_multiple_transform_to_categorical():
    cells = test_module.CellCollection()
    # y-type must be selected as a categorical candidate
    cells.properties['y-type'] = ['ugly', 'ugly', 'ugly', 'ugly', 'pretty']
    cells.properties['z-type'] = ['red', 'blue', 'green', 'alpha', 'optimus_prime']
    # y-type is a string at the beginning
    assert cells.properties['y-type'].dtypes == object
    assert cells.properties['z-type'].dtypes == object
    restored = check_roundtrip(cells)
    restored.properties['y-type'].to_frame()
    # y-type should be categorical now
    assert is_cat(restored.properties['y-type'])
    assert cells.properties['z-type'].dtypes == object
    restored.properties['z-type'].to_frame()


def test_roundtrip_positions():
    cells = test_module.CellCollection()
    cells.positions = random_positions(10)
    check_roundtrip(cells)


def test_roundtrip_orientations():
    cells = test_module.CellCollection()
    cells.orientations = random_orientations(10)
    check_roundtrip(cells)


def test_roundtrip_complex():
    cells = test_module.CellCollection()
    n = 5

    cells.positions = random_positions(n)
    cells.orientations = random_orientations(n)
    cells.properties['all_none'] = pd.Categorical.from_codes(codes=[0] * n, categories=[''])
    cells.properties['synapse_class'] = pd.Categorical.from_codes(
        codes=[0, 1, 1, 0, 0], categories=['', 'INH'])
    cells.properties['mtype'] = ['L5_NGC', 'L5_BTC', 'L5_BTC', 'L6_LBC', 'L6_LBC']
    cells.properties['etype'] = ['cADpyr', 'dNAC', 'dNAC', 'bSTUT', 'bSTUT']
    cells.properties['morphology'] = [
        'dend-C250500A-P3_axon-C190898A-P2_-_Scale_x1.000_y1.025_z1.000_-_Clone_2',
        'C240300C1_-_Scale_x1.000_y0.975_z1.000_-_Clone_55',
        'C240300C1_-_Scale_x1.000_y0.975_z1.000_-_Clone_55',
        'dend-Fluo15_right_axon-Fluo2_right_-_Clone_37',
        'dend-Fluo15_right_axon-Fluo2_right_-_Clone_37',
    ]
    prefix = test_module.CellCollection.SONATA_DYNAMIC_PROPERTY
    cells.properties[prefix + 'current'] = np.arange(n)
    cells.properties[prefix + 'some_prop'] = np.random.choice(['t1', 't2'], n)
    check_roundtrip(cells)


def test_remove_unassigned_1():
    cells = test_module.CellCollection()
    n = 5
    positions = random_positions(n)
    orientations = random_orientations(n)
    cells.positions = positions
    cells.orientations = orientations
    cells.properties = pd.DataFrame({
        'foo': ['', 'a', None, 'b', 'c'],
        'bar': [0., None, 2., 3., 4.]
    })
    cells.remove_unassigned_cells()
    assert_array_equal(cells.positions, positions[[0, 3, 4]])
    assert_array_equal(cells.orientations, orientations[[0, 3, 4]])
    assert_frame_equal(
        cells.properties,
        pd.DataFrame({
            'foo': ['', 'b', 'c'],
            'bar': [0., 3., 4.]
        })
    )


def test_remove_unassigned_2():
    cells = test_module.CellCollection()
    n = 2
    cells.positions = random_positions(n)
    cells.orientations = random_orientations(n)
    cells.properties = pd.DataFrame({
        'foo': ['a', 'b'],
        'bar': [0, 1],
    })
    cells.remove_unassigned_cells()
    assert len(cells.positions) == n
    assert len(cells.orientations) == n
    assert len(cells.properties) == n


def test_remove_unassigned_3():
    cells = test_module.CellCollection()
    cells.properties = pd.DataFrame({
        'foo': ['a', None],
    })
    cells.remove_unassigned_cells()
    assert cells.positions is None
    assert cells.orientations is None


@pytest.mark.parametrize(
    ("index_offset", "df_index"),
    [
        (0, [0, 1, 2]),
        (1, [1, 2, 3]),
    ]
)
def test_as_dataframe(index_offset, df_index):
    cells = test_module.CellCollection()
    cells.positions = np.random.random((3, 3))
    cells.orientations = random_orientations(3)
    cells.properties['foo'] = np.array(['a', 'b', 'c'])
    df = cells.as_dataframe(index_offset=index_offset)

    assert sorted(df.columns) == ['foo', 'orientation', 'x', 'y', 'z']
    assert_array_equal(df['x'], cells.positions[:, 0])
    assert_array_equal(np.stack(df['orientation']), cells.orientations)
    assert_array_equal(df['foo'].values, cells.properties['foo'].values)

    # check that dataframe is indexed by GIDs
    assert_array_equal(df.index.values, df_index)

    # check that data is copied
    df['foo'] = ['q', 'w', 'v']
    assert_array_equal(cells.properties['foo'], ['a', 'b', 'c'])

    assert df.columns.inferred_type in ('string', 'unicode')


def test_size():
    # Nothing
    cells = test_module.CellCollection()
    assert cells.size() == 0
    assert len(cells) == cells.size()

    # positions only
    cells = test_module.CellCollection()
    cells.positions = np.random.random((3, 3))
    assert cells.size() == 3
    assert len(cells) == cells.size()

    # orientations only
    cells = test_module.CellCollection()
    cells.orientations = random_orientations(3)
    assert cells.size() == 3
    assert len(cells) == cells.size()

    # properties only
    cells = test_module.CellCollection()
    cells.properties['foo'] = np.array(['a', 'b', 'c'])
    assert cells.size() == 3
    assert len(cells) == cells.size()

    cells = test_module.CellCollection()
    cells.positions = np.random.random((3, 3))
    cells.orientations = random_orientations(3)
    cells.properties['foo'] = np.array(['a', 'b', 'c'])
    assert cells.size() == 3
    assert len(cells) == cells.size()

    # bad sizes : properties too small
    cells = test_module.CellCollection()
    cells.positions = np.random.random((3, 3))
    cells.orientations = random_orientations(3)
    cells.properties['foo'] = np.array(['a', 'b'])
    with pytest.raises(VoxcellError):
        cells.size()


def test_add_properties():
    cells = test_module.CellCollection()
    properties1 = pd.DataFrame({
        'a': [1],
        'b': [2],
    })
    properties2 = pd.DataFrame({
        'b': [3],
        'c': [4],
    })
    combined = pd.DataFrame({
        'a': [1],
        'b': [3],
        'c': [4],
    })

    cells.add_properties(properties1)
    assert_frame_equal(cells.properties, properties1)

    # no duplicates should appear
    cells.add_properties(properties2)
    assert_frame_equal(cells.properties, combined)

    # no overwriting => exception should be raised if column already exists
    with pytest.raises(VoxcellError):
        cells.add_properties(properties1, overwrite=False)


@pytest.mark.parametrize(
    ("index_offset", "df_index"),
    [
        (1, [0, 1]),
        (0, [1, 2]),
    ]
)
def test_from_dataframe_invalid_index(index_offset, df_index):
    df = pd.DataFrame({
        'prop-a': ['a', 'b'],
    }, index=df_index)
    with pytest.raises(VoxcellError, match="Index !="):
        test_module.CellCollection.from_dataframe(df, index_offset=index_offset)


@pytest.mark.parametrize(
    ("index_offset", "df_index"),
    [
        (0, [0, 1]),
        (1, [1, 2]),
    ]
)
def test_from_dataframe_no_positions(index_offset, df_index):
    df = pd.DataFrame({
        'prop-a': ['a', 'b'],
    }, index=df_index)

    cells = test_module.CellCollection.from_dataframe(df, index_offset=index_offset)
    assert cells.positions is None
    assert cells.orientations is None
    assert_frame_equal(cells.properties, df.reset_index(drop=True))


@pytest.mark.parametrize("index_offset", [0, 1])
def test_to_from_dataframe(index_offset):
    cells = test_module.CellCollection()
    cells.positions = random_positions(3)
    cells.orientations = random_orientations(3)
    cells.properties['foo'] = np.array(['a', 'b', 'c'])
    cells.properties['cat'] = pd.Categorical.from_codes(
        codes=np.zeros(3, dtype=np.uint), categories=['a'])

    df = cells.as_dataframe(index_offset=index_offset)
    cells2 = test_module.CellCollection.from_dataframe(df, index_offset=index_offset)
    assert_almost_equal(cells.positions, cells2.positions)
    assert_almost_equal(cells.orientations, cells2.orientations)
    assert_frame_equal(cells.properties, cells2.properties)


def assert_sonata_rotation(filename, good_orientation_type):
    cells = test_module.CellCollection.load_sonata(filename)
    assert cells.orientation_format == good_orientation_type
    check_roundtrip(cells)
    # cannot be included in the roundtrip because the mvd3 conversion will potentially break
    # the format
    with tempcwd():
        original = test_module.CellCollection.load_sonata(filename)
        original.save_sonata("nodes_.h5")
        restored = test_module.CellCollection.load_sonata("nodes_.h5")
        assert original.orientation_format == restored.orientation_format
        for axis in ["x", "y", "z", "w"]:
            assert "rotation_angle_{}axis".format(axis) not in restored.properties
            assert "rotation_angle_{}axis".format(axis) not in original.properties
            assert "orientation_{}".format(axis) not in restored.properties
            assert "orientation_{}".format(axis) not in original.properties


def test_load_sonata_orientations():
    assert_sonata_rotation(SONATA_DATA_PATH / "nodes_eulers.h5", "eulers")
    assert_sonata_rotation(SONATA_DATA_PATH / "nodes_quaternions.h5", "quaternions")
    assert_sonata_rotation(SONATA_DATA_PATH / "nodes_no_rotation.h5", "quaternions")
    with pytest.raises(VoxcellError):
        assert_sonata_rotation(SONATA_DATA_PATH / "nodes_quaternions_w_missing.h5", "quaternions")


def test_set_orientation_type():
    with tempcwd():
        cells = test_module.CellCollection.load_sonata(SONATA_DATA_PATH / "nodes_eulers.h5")
        assert cells.orientation_format == "eulers"
        cells.orientation_format = "quaternions"
        cells.save_sonata("nodes_.h5")
        restored = test_module.CellCollection.load_sonata("nodes_.h5")
        assert restored.orientation_format == "quaternions"
        for axis in ["x", "y", "z"]:
            assert "rotation_angle_{}axis".format(axis) not in restored.properties

        with pytest.raises(VoxcellError):
            cells.orientation_format = "unknown"


def test_sonata_multipopulation():
    with tempcwd():
        orig = test_module.CellCollection.load_sonata(SONATA_DATA_PATH / "nodes_eulers.h5")

        orig.population_name = 'A'
        orig.save_sonata("nodes.h5")

        orig.population_name = 'B'
        # this tests that `mode` is used, otherwise the file would be overwritten, and population
        # A wouldn't be accessible below
        orig.save_sonata("nodes.h5", mode='a')

        A = test_module.CellCollection.load_sonata("nodes.h5", population_name="A")
        B = test_module.CellCollection.load_sonata("nodes.h5", population_name="B")
        assert_frame_equal(A.as_dataframe(), B.as_dataframe())
        assert_frame_equal(A.as_dataframe(index_offset=0), B.as_dataframe(index_offset=0))
        assert_frame_equal(A.as_dataframe(index_offset=1), B.as_dataframe(index_offset=1))


def test_check_types():
    # this is a sanity check for the h5py>3.0.0 and the string types
    cells = test_module.CellCollection.load_sonata(SONATA_DATA_PATH / "nodes_multi_types.h5")
    assert_array_equal(cells.properties["string"],
                       ['AA', 'BB', 'CC', 'DD', 'EE', 'FF', 'GG'])
    assert_array_equal(cells.properties["categorical"], ['A', 'A', 'B', 'A', 'A', 'A', 'A'])
    assert_array_equal(cells.properties["int"], [0, 0, 1, 0, 0, 0, 0])
    assert_almost_equal(cells.properties["float"], [0.0, 0.0, 1.1, 0.0, 0.0, 0.0, 0.0])
    assert cells.properties["float"].to_numpy().dtype == np.float32


def test_save_sonata_forced_library():
    with tempcwd():
        orig = test_module.CellCollection.load_sonata(SONATA_DATA_PATH / "nodes_multi_types.h5")
        orig.save_sonata("nodes_.h5", forced_library=["categorical", "string"])
        with h5py.File("nodes_.h5", "r") as h5:
            assert list(h5['/nodes/default/0/@library'].keys()) == ["categorical", "string"]
        restored = test_module.CellCollection.load_sonata("nodes_.h5")
        assert_equal_cells(orig, restored)


def test_str_and_repr():
    for op in (str, repr):
        cc = test_module.CellCollection.load_sonata(SONATA_DATA_PATH / "nodes_quaternions.h5")
        ret = op(cc)
        assert 'orientation_x' in ret
        assert 'orientation_y' in ret
        assert 'orientation_z' in ret
        assert 'orientation_w' in ret

        cc = test_module.CellCollection.load_sonata(SONATA_DATA_PATH / "nodes_eulers.h5")
        ret = op(cc)
        assert 'rotation_angle_xaxis' in ret
        assert 'rotation_angle_yaxis' in ret
        assert 'rotation_angle_zaxis' in ret

        cc = test_module.CellCollection.load_sonata(SONATA_DATA_PATH / "nodes_multi_types.h5")
        ret = op(cc)
        assert 'categorical' in ret
        assert 'float' in ret
        assert 'string' in ret
        assert 'int' in ret



"""Library to build, transform and handle fields of vectors and orientations.

A vector field is a volumetric dataset that expresses a 3D vector for each voxel.
Vector fields are represented as 4D numpy arrays. The first three dimensions of
the array represent space, the last dimension is always of size 3 and contains the three components
(i, j, k) of the vector for that voxel.


An orientation field is a volumetric dataset that expresses a rotation matrix for each voxel.
Orientation fields are represented as 5D numpy arrays. The first three dimensions of the array
represent space, the fourth and fifth dimensions contain a 3x3 rotation matrix.

Note that although 3D is our main target, most of these functions should behave correctly for lower
and higher dimensionality levels.
"""

import logging

import numpy as np
import scipy.ndimage

L = logging.getLogger(__name__)


def gaussian_filter(vf, sigma):
    """Apply a gaussian filter to a vector field.

    Note that filter is applied without normalizing the input or output.

    Args:
        vf: Vector field
        sigma : scalar. Standard deviation for Gaussian kernel.

    Returns:
        The resulting vector field not normalized.
    """
    filtered = vf.copy()
    mask = np.any(vf != 0, axis=-1)

    filtered[..., 0] = mask * scipy.ndimage.gaussian_filter(filtered[..., 0], sigma=sigma)
    filtered[..., 1] = mask * scipy.ndimage.gaussian_filter(filtered[..., 1], sigma=sigma)
    filtered[..., 2] = mask * scipy.ndimage.gaussian_filter(filtered[..., 2], sigma=sigma)

    return filtered


def combine_vector_fields(fields):
    """Given a list of vector fields return an orientation field.

    The vectors from the fields are treated as a new base and will be stored as column vectors so
    that the matrices of the resulting field (one per voxel) can be used to premultiply vectors
    for rotation.

    Args:
        fields: a list of vector fields.
            All of the fields are expected to have the same shape (AxBxCx3)

    Returns:
        A 5D numpy array representing an orientation field
    """
    if not fields:
        return np.empty((0,))

    shape = fields[0].shape

    # add a second-to-last dimension: the number of fields
    result = np.zeros(shape=shape[:-1] + (len(fields),) + (shape[-1],), dtype=fields[0].dtype)

    # abusing numpy broadcasting here saves us having to do an explicit transpose afterwards
    for i, f in enumerate(fields):
        result[..., i] = f

    return result


def split_orientation_field(field):
    """Given an orientation field return a list of vector fields."""
    return [field[..., i] for i in range(field.shape[-2])]


def join_vector_fields(vf0, *vfs):
    """Performs an union of several vector fields.

    A voxel on a vector field is considered "empty" if all of its components are zero.

    Args:
        vf0: first vector field.
        vfs: rest of vector fields. All of them must have the same shape.
    """
    vfs = (vf0,) + vfs
    assert all(v.shape == vf0.shape for v in vfs)  # pylint: disable=no-member

    joined = np.zeros_like(vf0)
    joined_mask = np.zeros(joined.shape[:-1],  # pylint: disable=unsubscriptable-object
                           dtype=bool)
    for field in vfs:
        field_mask = np.any(field != 0, axis=-1)
        overlap_count = np.count_nonzero(joined_mask & field_mask)
        if overlap_count:
            L.warning('%d voxels overlap will be overwritten', overlap_count)
        joined_mask |= field_mask
        joined[field_mask] = field[field_mask]

    return joined



"""Library to handle traits fields and collections and the logic to assign them.

A "trait" is a set of properties and their values.
A "traits collection" is a group of traits. It's represented as a pandas
DataFrame object where each trait is a row and each property is a column.
For example::

       sclass      mtype
    0  INH         Pyramidal
    1  EXC         Pyramidal
    2  INH         Martinotti

A "probability distribution" specifies the probability of each trait in a collection
being assigned to a given cell.
A "distributions collection" is a group of distributions. It's represented
as a pandas DataFrame object where each trait is a row and each distribution
is a column.
For example::

        0     1    2     3
     0  0.25  0.5  0.5   0.0
     1  0.75  0.5  0.25  0.1
     2  0.0   0.0  0.25  0.9

The distribution number 3 in the above table indicates 10% chances of picking
the trait 1 (an excitatory Pyramidal cell) and 90% of picking trait 2
(an inhibitory Martinotti cell).
"""

import logging

import numpy as np
import pandas as pd

L = logging.getLogger(__name__)


class SpatialDistribution:
    """Encapsulates the data relative to the 3D probability distributions of traits.

    Args:
        field: VoxelData where every voxel is an index in the distributions list
               -1 means unavailable data.
        distributions: a distributions collection, see module docstring
        traits: a traits collection, see module docstring
    """

    def __init__(self, field, distributions, traits):
        """Init SpatialDistribution."""
        self.field = field
        self.distributions = distributions / distributions.sum()
        self.traits = traits

    def assign(self, positions):
        """For every cell in positions, chooses a property from a spatial distribution.

        Args:
            positions: list of positions for soma centers (x, y, z).

        Returns:
            An array with the same length as positions where each value
            is an index into spatial_dist.traits
        """
        dist_id_per_position = self.field.lookup(positions)

        unknown_count = np.count_nonzero(dist_id_per_position == -1)
        if unknown_count:
            L.warning('%d total positions in unknown areas', unknown_count)

        valid = np.in1d(dist_id_per_position, self.distributions.columns)
        if np.count_nonzero(~valid):
            L.warning('missing distribution for %d positions', np.count_nonzero(~valid))

        chosen_trait_indices = np.ones_like(dist_id_per_position, dtype=np.int32) * -1

        unique_dists = np.unique(dist_id_per_position[valid])
        for dist_id, dist in self.distributions[unique_dists].items():
            hit_count = np.count_nonzero(dist_id_per_position == dist_id)
            chosen = np.random.choice(dist.index, hit_count, p=dist.values)
            chosen_trait_indices[dist_id_per_position == dist_id] = chosen

        return chosen_trait_indices

    def assign_conditional(self, positions, preassigned):
        """For every cell in positions, choose a property from a spatial distribution.

        For every cell in positions, choose a property from a spatial distribution,
        but taking into account a pre-assigned values of related properties.

        Args:
            positions: list of positions for soma centers (x, y, z).
            preassigned: pandas.DataFrame or pandas.Series with the pre-assigned values.

        Returns:
            An array with the same length as positions where each value
            is an index into spatial_dist.traits

            For those positions whose value could not be determined, -1 is used.
        """
        # No need to apply offset here because it's applied inside subdist.assign
        preassigned = preassigned.to_frame() if isinstance(preassigned, pd.Series) else preassigned

        subsections = self.split(list(preassigned.columns))
        chosen = np.ones(shape=(len(preassigned)), dtype=int) * -1

        unique_assigned = preassigned.drop_duplicates()
        for values_comb in unique_assigned.values:
            hashable = tuple(values_comb)

            if hashable in subsections:
                subdist = subsections[hashable]
                mask = np.all(np.array(preassigned) == values_comb, axis=1)
                chosen[mask] = subdist.assign(positions[mask])

        invalid_count = np.count_nonzero(chosen == -1)
        if invalid_count:
            # this may happen because of inconsistencies of the data
            # for example if we assigned excitatory neurons to a neuron that is in
            # a voxel for which only inhibitory metype probabilities are known
            L.warning('%d / %d = %f cells could not get a valid value assigned',
                      invalid_count, len(chosen),
                      float(invalid_count) / len(chosen))

        return chosen

    def collect(self, positions, preassigned=None, names=None):
        """For every cell in positions, chose trait values from a spatial distribution.

        For every cell in positions, chose trait values from a spatial distribution,
        taking into account a pre-assigned values of related properties (if specified).

        Args:
            positions: list of positions for soma centers (x, y, z).
            preassigned: pandas.DataFrame or pandas.Series with the pre-assigned values.
            names: names of the properties to collect. If not specified use all in traits.

        Returns:
            A pandas DataFrame with one row for each index and one column for each value of names
        """
        if preassigned is None:
            return self.collect_traits(self.assign(positions), names)

        return self.collect_traits(self.assign_conditional(positions, preassigned), names)

    def collect_traits(self, chosen, names=None):
        """Return the trait values corresponding to an array of indices.

        Args:
            chosen: array of indices into the traits dataframe
            names: names of the properties to collect. If not specified use all in traits.

        Returns:
            A pandas DataFrame with one row for each index and one column for each value of names
        """
        names = names if names is not None else self.traits.columns
        df = self.traits[names].iloc[chosen]
        return df.reset_index().drop('index', axis=1)

    def split(self, attributes):
        """Split a distribution in two or more.

        Split a distribution in two or more so that each one only references
        traits with the same value for certain attributes.
        Each resulting distribution is renormalised.

        this may be generating distributions that are empty but that should not be a problem

        Note that because for every distribution we are creating a new one,
        the indexes of any associated field are still valid.

        Returns:
            A dictionary where the keys are tuples with the values of the attributes found
            in the traits_collection and the values are the resulting SpatialDistribution objects.
        """
        grouped_distributions = {}
        for attr_values, traits in self.traits.groupby(attributes):
            # this is backwards compatibility: https://github.com/pandas-dev/pandas/issues/42795
            if len(attributes) == 1:
                attr_values = tuple(attr_values)

            dists = self.distributions.iloc[traits.index]

            # remove dists that have become empty
            dists = dists[dists.columns[dists.sum() != 0]]
            dists /= dists.sum()

            grouped_distributions[attr_values] = SpatialDistribution(self.field, dists, traits)

        return grouped_distributions

    def reduce(self, attribute):
        """Return a SpatialDistribution with only the given attribute.

        Given a spatial distribution, extract an equivalent one where all of the properties
        of the traits collection have been removed except for a specific one.

        For example, taking the traits collection::

               sclass      mtype
            0  INH         Pyramidal
            1  EXC         Pyramidal
            2  INH         Martinotti

        and the distributions collection::

               0
            0  0.2
            1  0.4
            2  0.4

        Ignoring all properties except 'sclass' would give the simplified traits collection::

               sclass
            0  INH
            1  EXC

        and the distributions collection::

               0
            0  0.6
            1  0.4

        Note that because for every distribution we are creating a new one,
        the indexes of any associated field are still valid.

        Returns:
            An SpatialDistribution object
        """
        traits, distributions = [], []
        for value, data in self.traits.groupby(attribute):
            traits.append(value)
            distributions.append(self.distributions.iloc[data.index].sum())

        traits = pd.DataFrame(traits, columns=[attribute])
        distributions = pd.DataFrame(
            distributions,
            columns=self.distributions.columns,
            index=traits.index
        )

        return SpatialDistribution(self.field, distributions, traits)

    def get_probability_field(self, attribute, value):
        """Extract the probability of a particular attribute value from a spatial distribution.

        Voxels where the probability is missing will contain a probability value of -1.

        Returns:
            A VoxelData with the same shape as sdist.field where every voxel is a float
            representing the probability of an attribute value being assigned to a cell
            in that voxel. For voxels with unknown values -1 is used.
        """
        probs = self.distributions[self.traits[attribute] == value].sum()
        probs_field = probs.reindex(self.field.raw.flatten())
        probs_field = probs_field.fillna(-1)
        return self.field.with_data(probs_field.values.reshape(self.field.raw.shape))

    @classmethod
    def from_probability_field(cls, probs_field, name, positive_value, negative_value):
        """Creates a binary SpatialDistribution object from a single probability field.

        Args:
            probs_field: A VoxelData where every voxel is a float representing the probability
                of an attribute value being assigned to a cell in that voxel. For voxels with
                unknown values -1 is used.
            name: name of the trait.
            positive_value: value of the trait that corresponds to probs_field.
            negative_value: value of the trait that corresponds to the probability field
                complementary to probs_field.

        Returns:
            A SpatialDistribution object
        """
        valid = probs_field.raw != -1

        probs_field_complete = probs_field.raw.copy()
        any_value = probs_field.raw[valid].flatten()[0]
        probs_field_complete[~valid] = any_value
        unique_probs, field = np.unique(probs_field_complete, return_inverse=True)

        field = field.reshape(probs_field.raw.shape)
        field[~valid] = -1
        field = probs_field.with_data(field)

        distributions = pd.DataFrame([unique_probs, 1 - unique_probs])
        traits = pd.DataFrame({name: [positive_value, negative_value]})

        return cls(field, distributions, traits)

    def drop_duplicates(self, decimals=None):
        """Return a new SpatialDistribution with duplicate distributions removed.

        Args:
            decimals: Number of decimal places to round each column to beforehand.
        """
        dists, inverse = _drop_duplicate_columns(self.distributions, decimals)

        field = np.ones_like(self.field.raw) * -1
        for col_idx, unique_idx in enumerate(inverse):
            field[self.field.raw == col_idx] = unique_idx

        return SpatialDistribution(self.field.with_data(field), dists, self.traits.copy())


def _drop_duplicate_columns(dataframe, decimals=None):
    """Drop duplicate columns from a dataframe and return the indices of the columns.

    Drop duplicate columns from a dataframe and return the indices of the columns
    of the resulting dataframe that can be used to reconstruct the original one.

    Args:
        dataframe: pandas.Dataframe object
        decimals: Number of decimal places to round each column to beforehand.
    """
    df = dataframe.transpose()

    if decimals is not None:
        df = df.round(decimals)

    g = df.groupby(list(df.columns))
    df1 = df.set_index(list(df.columns))
    tags = df1.index.map(lambda ind: g.indices[ind][0])
    unique_rows_idx, inverse = np.unique(tags, return_inverse=True)

    return df.iloc[unique_rows_idx].transpose(), inverse



import json
import os
from unittest.mock import mock_open, patch

import pytest

import voxcell.region_map as test_module
from voxcell.exceptions import VoxcellError

DATA_PATH = os.path.join(os.path.dirname(__file__), 'data')

TEST_RMAP_DICT = {
    'id': 0,
    'name': 'root',
    'fullname': 'The Root Node',
    'children': [
        {
        'id': 1,
        'name': 'A',
        'fullname': 'aA',
        'children': [
            {
                'id': 2,
                'name': 'B',
                'fullname': 'Bb',
                'children': [],
            },
            {
                'id': 3,
                'name': 'C',
                'fullname': 'cC',
                'children': [
                    {
                        'id': 4,
                        'name': 'B',
                        'fullname': 'bB',
                        'children': [],
                    }
                ]
            }
        ]}
    ]
}

TEST_RMAP = test_module.RegionMap.from_dict(TEST_RMAP_DICT)


def test_find_basic():
    assert TEST_RMAP.find('B', 'name') == {2, 4}
    assert TEST_RMAP.find('C', 'name') == {3}
    assert TEST_RMAP.find('cC', attr='fullname') == {3}


def test_find_regex():
    assert TEST_RMAP.find('B', attr='fullname') == set()
    assert TEST_RMAP.find('@B', attr='fullname') == {2, 4}
    assert TEST_RMAP.find('@^B', attr='fullname') == {2}
    assert TEST_RMAP.find('@B$', attr='fullname') == {4}


def test_find_with_descendants():
    assert TEST_RMAP.find('B', 'name', with_descendants=True) == {2, 4}
    assert TEST_RMAP.find('C', 'name', with_descendants=True) == {3, 4}


def test_find_ignore_case():
    assert TEST_RMAP.find('cc', attr='fullname') == set()
    assert TEST_RMAP.find('cc', attr='fullname', ignore_case=True) == {3}
    assert TEST_RMAP.find('CC', attr='fullname', ignore_case=True) == {3}

def test_find_regex_ignore_case():
    assert TEST_RMAP.find('@^B', attr='fullname') == {2}
    assert TEST_RMAP.find('@^B', attr='fullname', ignore_case=True) == {2, 4}


def test_find_id():
    assert TEST_RMAP.find(1, attr='id', with_descendants=True) == {1, 2, 3, 4}
    assert TEST_RMAP.find(1, attr='id', ignore_case=True) == {1}
    assert TEST_RMAP.find(999, attr='id') == set()


def test_find_missing_attribute():
    with pytest.raises(VoxcellError):
        TEST_RMAP.find(1, attr='no-such-attribute')


def test_get_basic():
    assert TEST_RMAP.get(4, 'name') == 'B'
    assert TEST_RMAP.get(4, attr='fullname') == 'bB'


def test_get_with_ascendants():
    assert TEST_RMAP.get(4, 'name', with_ascendants=True) == ['B', 'C', 'A', 'root']


def test_get_missing_attribute():
    with pytest.raises(VoxcellError):
        TEST_RMAP.get(4, 'no-such-attribute')


def test_get_missing_id():
    with pytest.raises(VoxcellError):
        TEST_RMAP.get(999, 'name')


def _patch_open_json(data):
    return patch('voxcell.region_map.open', new_callable=mock_open, read_data=json.dumps(data))


def test_load_json_basic():
    data = {
        'id': 42,
        'name': "foo",
    }
    with _patch_open_json(data):
        rmap = test_module.RegionMap.load_json('mock-file')
    assert rmap.get(42, 'name') == "foo"


def test_load_json_aibs():
    data = {
        'msg': [
            {
                'id': 42,
                'name': "foo",
            },
        ]
    }
    with _patch_open_json(data):
        rmap = test_module.RegionMap.load_json('mock-file')
    assert rmap.get(42, 'name') == "foo"


def test_load_json_aibs_raises():
    data = {
        'msg': [
            {
                'id': 42,
                'name': "foo",
            },
            {
                'id': 43,
                'name': "bar",
            }
        ]
    }
    with _patch_open_json(data), pytest.raises(VoxcellError):
        test_module.RegionMap.load_json('mock-file')


def test_from_dict_duplicate_id():
    with pytest.raises(VoxcellError):
        test_module.RegionMap.from_dict({
            'id': 1,
            'children': [
                {
                    'id': 2
                },
                {
                    'id': 1
                }
            ]
        })


def test_as_dict():
    res = TEST_RMAP.as_dict()
    assert res == TEST_RMAP_DICT

    with open(os.path.join(DATA_PATH, "region_map.json")) as f:
        initial_dict = json.load(f)
    rmap = test_module.RegionMap.from_dict(initial_dict)
    assert rmap.as_dict() == initial_dict


def test_is_leaf_id():
    assert TEST_RMAP.is_leaf_id(1) is False
    assert TEST_RMAP.is_leaf_id(2) is True
    assert TEST_RMAP.is_leaf_id(3) is False
    assert TEST_RMAP.is_leaf_id(4) is True


def test_is_leaf_id_non_existing_id():
    with pytest.raises(VoxcellError):
        TEST_RMAP.is_leaf_id(9999)  # non-existing id


def test_as_dataframe():
    df = TEST_RMAP.as_dataframe()
    assert df.loc[0].parent_id == -1
    assert df.loc[1].parent_id == 0
    assert df.loc[2].parent_id == 1
    assert df.loc[3].parent_id == 1
    assert df.loc[4].parent_id == 3

    assert df.loc[1]['name'] == 'A'
    assert df.loc[1]['fullname'] == 'aA'

    assert df.loc[0].children_count == 1
    assert df.loc[1].children_count == 2
    assert df.loc[2].children_count == 0
    assert df.loc[3].children_count == 1
    assert df.loc[4].children_count == 0


def test_from_dataframe():
    # Test with a simple RegionMap
    rmap = test_module.RegionMap.from_dataframe(TEST_RMAP.as_dataframe())
    assert rmap._data == TEST_RMAP._data
    assert rmap._parent == TEST_RMAP._parent
    assert rmap._children == TEST_RMAP._children

    # Test with more complex data
    initial_rmap = test_module.RegionMap.load_json(os.path.join(DATA_PATH, "region_map.json"))
    final_rmap = test_module.RegionMap.from_dataframe(initial_rmap.as_dataframe())
    assert final_rmap._data == initial_rmap._data
    assert final_rmap._parent == initial_rmap._parent
    assert final_rmap._children == initial_rmap._children

    # Test with multiple root nodes
    rmap_df = initial_rmap.as_dataframe()
    rmap_df.loc[8, "parent_id"] = -1
    with pytest.raises(RuntimeError):
        test_module.RegionMap.from_dataframe(rmap_df)



import operator
import os
import re
import tempfile
from pathlib import Path
from unittest.mock import Mock, call

import nrrd
import numpy as np
import numpy.testing as npt
import pytest
from numpy.testing import assert_almost_equal, assert_array_equal, assert_raises

import voxcell.voxel_data as test_module
from voxcell.exceptions import VoxcellError

DATA_PATH = os.path.join(os.path.dirname(__file__), 'data')


def test_lookup():
    raw = np.array([[11, 12], [21, 22]])
    v = test_module.VoxelData(raw, (2, 3), offset=np.array([2, 2]))
    assert v.lookup([[2, 3]]) == [11]
    assert_array_equal(v.lookup([[2, 10], [1, 5]], outer_value=42), [42, 42])
    assert_array_equal(v.lookup([[2, 10], [1, 5], [4, 5]], outer_value=42), [42, 42, 22])
    assert_raises(VoxcellError, v.lookup, [[1, 5]])
    assert_raises(VoxcellError, v.lookup, [[2, 10]])


@pytest.mark.parametrize(
    "raw, voxel_dimensions, coordinates, outer_value, expected",
    [
        (
            [[[11, 11], [12, 12]], [[21, 21], [22, 22]]],
            (2, 2),
            [[1, 1], [3, 3]],
            None,
            [[11, 11], [22, 22]],
        ),
        (
            [[[11, 11], [12, 12]], [[21, 21], [22, 22]]],
            (2, 2),
            [[10, 1]],
            [10, 10],
            [[10, 10]],
        ),
        (
            np.zeros((1, 1, 1)),
            (1, 1, 1),
            [(-1, -1, -1)],
            0,
            [0],
        ),
        (
            np.zeros((1, 1, 1, 1)),
            (1, 1, 1),
            [(-1, -1, -1)],
            0,
            [(0,)],
        ),
        (
            np.zeros((1, 1, 1, 2)),
            (1, 1, 1),
            [(-1, -1, -1)],
            (0, 0),
            [(0, 0)],
        ),
    ],
)
def test_lookup_vector_data(raw, voxel_dimensions, coordinates, outer_value, expected):
    voxel_data = test_module.VoxelData(np.array(raw), voxel_dimensions)
    assert_array_equal(voxel_data.lookup(coordinates, outer_value), expected)


def test_positions_to_indices():
    raw = np.zeros(2)
    v = test_module.VoxelData(raw, voxel_dimensions=(10.0,), offset=(10.0,))
    assert_array_equal(v.positions_to_indices([15., 25.]), [0, 1])
    assert_array_equal(v.positions_to_indices([15., 25.], keep_fraction=True), [0.5, 1.5])

    # border effects
    assert v.positions_to_indices([9.9999999]) == [0]
    assert v.positions_to_indices([19.9999999]) == [0]

    # messing up with rounding errors
    v = test_module.VoxelData(raw, voxel_dimensions=(1.125,), offset=(-0.5,))
    assert_raises(VoxcellError, v.positions_to_indices, [-0.5, 1.75])
    assert_array_equal(v.positions_to_indices([-0.5, 1.75 - 1e-15]), [0, 1])
    assert_raises(VoxcellError, v.positions_to_indices, [-0.5, 1.75], keep_fraction=True)

    assert_array_equal(
        v.positions_to_indices([-0.5, ], keep_fraction=False),
        [0, ],
    )

    # w/ voxel_dimensions=(1.125,), offset=(-0.5,), on_edge will IEEE 'round-to-even'
    # which means that it is considered out of bounds, since we do exclusive ranges
    # on the far edges, which is incorrect
    on_edge = np.nextafter(1.75, -1, dtype=float)
    assert_array_equal(
        v.positions_to_indices([on_edge, ], keep_fraction=False),
        [1, ],
    )

    assert_array_equal(
        v.positions_to_indices([-0.5, on_edge, ], keep_fraction=True),
        [0., np.nextafter(2, -1, dtype=float)],
    )


def test_load_nrrd_scalar_payload():
    actual = test_module.VoxelData.load_nrrd(os.path.join(DATA_PATH, 'scalar.nrrd'))
    assert actual.raw.shape == (1, 2)
    assert_almost_equal(actual.voxel_dimensions, [10, 20])
    assert_almost_equal(actual.offset, [100, 200])
    assert_almost_equal(actual.bbox, np.array([[100, 200], [110, 240]]))


def test_load_nrrd_vector_payload():
    actual = test_module.VoxelData.load_nrrd(os.path.join(DATA_PATH, 'vector.nrrd'))
    assert actual.raw.shape == (1, 2, 3)
    assert_almost_equal(actual.raw, [[[11, 12, 13], [21, 22, 23]]])
    assert_almost_equal(actual.voxel_dimensions, [10, 20])
    assert_almost_equal(actual.offset, [100, 200])


def test_load_nrrd_with_space_directions():
    actual = test_module.VoxelData.load_nrrd(os.path.join(DATA_PATH, 'space_directions.nrrd'))
    assert actual.raw.shape == (1, 2, 3)
    assert_almost_equal(actual.voxel_dimensions, [10, 20])
    assert_almost_equal(actual.offset, [100, 200])


def test_load_nrrd_fail():
    # no spacing information
    assert_raises(
        VoxcellError,
        test_module.VoxelData.load_nrrd,
        os.path.join(DATA_PATH, 'no_spacings_fail.nrrd'),
    )
    # space directions is non-diagonal
    assert_raises(
        NotImplementedError,
        test_module.VoxelData.load_nrrd,
        os.path.join(DATA_PATH, 'space_directions_fail.nrrd'),
    )


def test_save_nrrd():
    """Test saving a test nrrd file and check basic attributes."""
    vd = test_module.VoxelData.load_nrrd(os.path.join(DATA_PATH, 'vector.nrrd'))
    with tempfile.NamedTemporaryFile(suffix='.nrrd') as f:
        vd.save_nrrd(f.name)
        f.flush()
        f.seek(0)
        new = test_module.VoxelData.load_nrrd(f.name)
        assert np.allclose(vd.raw, new.raw)
        assert np.allclose(vd.voxel_dimensions, new.voxel_dimensions)
        assert np.allclose(vd.offset, new.offset)
    with tempfile.NamedTemporaryFile(suffix='.nrrd') as f:
        vd.save_nrrd(f.name, encoding='raw')
        f.flush()
        f.seek(0)
        new = test_module.VoxelData.load_nrrd(f.name)
        assert np.allclose(vd.raw, new.raw)


def test_save_load_nrrd_as_str():
    """Test saving and loading using a string file path"""
    vd = test_module.VoxelData.load_nrrd(os.path.join(DATA_PATH, 'vector.nrrd'))
    with tempfile.NamedTemporaryFile(suffix='.nrrd') as f:

        string = str(f.name)
        vd.save_nrrd(string)
        test_module.VoxelData.load_nrrd(string)


def test_save_load_nrrd_as_path():
    """Test saving and loading using a pathlib.Path"""
    vd = test_module.VoxelData.load_nrrd(os.path.join(DATA_PATH, 'vector.nrrd'))
    with tempfile.NamedTemporaryFile(suffix='.nrrd') as f:

        path = Path(f.name)
        vd.save_nrrd(path)
        test_module.VoxelData.load_nrrd(path)


def test_save_nrrd_with_extra_axes():
    """Test saving a numpy array with more than 3 dimensions."""
    raw = np.zeros((6,7,8,4,3)) # two extra dimensions
    vd = test_module.VoxelData(raw, (1.0, 2.0, 3.0))
    with tempfile.NamedTemporaryFile(suffix='.nrrd') as f:
        vd.save_nrrd(f.name)
        f.flush()
        f.seek(0)
        _, header = nrrd.read(f.name)
        # pynrrd will convert None into np.array([np.nan, np.nan, np.nan])
        assert_array_equal(header['space directions'],
            [
                [np.nan] * 3, [np.nan] * 3, (1.0, 0.0, 0.0), (0.0, 2.0, 0.0), (0.0, 0.0, 3.0)
            ])
        assert 'kinds' not in header


def test_save_nrrd_vector_field():
    """Test saving a numpy array with exactly 4 dimensions and a numeric dtype."""
    raw = np.zeros((6, 7, 8, 5)) # one extra dimension
    vd = test_module.VoxelData(raw, (1.0, 2.0, 3.0))
    with tempfile.NamedTemporaryFile(suffix='.nrrd') as f:
        vd.save_nrrd(f.name)
        f.flush()
        f.seek(0)
        _, header = nrrd.read(f.name)
        # pynrrd's reader will convert None into np.array([np.nan, np.nan, np.nan])
        assert_array_equal (header['space directions'],
            [
                [np.nan] * 3, (1.0, 0.0, 0.0), (0.0, 2.0, 0.0), (0.0, 0.0, 3.0)
            ])
        assert_array_equal(header['kinds'], ['vector', 'domain', 'domain', 'domain'])


def test_shape_checks():
    raw = np.zeros(shape=(2, 2))
    assert_raises(VoxcellError, test_module.VoxelData, raw, voxel_dimensions=[[25], [25]])
    assert_raises(VoxcellError, test_module.VoxelData, raw, voxel_dimensions=(25, 25), offset=(0,))
    assert_raises(VoxcellError, test_module.VoxelData, raw, voxel_dimensions=(25, 25, 25))


def test_with_data():
    original = test_module.VoxelData(np.array([0, 1]), voxel_dimensions=(2,), offset=(42,))
    replaced = original.with_data(np.array([2, 3]))
    assert_array_equal(replaced.raw, [2, 3])
    assert replaced.voxel_dimensions == original.voxel_dimensions
    assert replaced.offset == original.offset


def test_indices_to_positions():
    vd = test_module.VoxelData(np.array([0, 1]), voxel_dimensions=(2,), offset=(42,))
    positions = vd.indices_to_positions(np.array([0, 0.5, 1]))
    assert_almost_equal(positions, [42, 43, 44])


def test_count():
    vd = test_module.VoxelData(np.array([0, 1, 1, 2]), voxel_dimensions=(2,))
    assert vd.count(7) == 0
    assert vd.count(1) == 2
    assert vd.count([0, 2]) == 2
    assert vd.count(set([0, 2])) == 2


def test_volume():
    vd = test_module.VoxelData(np.array([[0, 1], [1, 2]]), voxel_dimensions=(2, -3))
    assert vd.volume(1) == 12
    assert vd.volume(13) == 0


def test_clip():
    raw = np.array([1, 2, 3])
    original = test_module.VoxelData(raw, voxel_dimensions=(2,), offset=(10,))
    clipped = original.clip(bbox=((11,), (15,)), na_value=-1)
    assert_array_equal(original.raw, raw)
    assert_array_equal(clipped.raw, [-1, 2, -1])
    assert clipped.voxel_dimensions == original.voxel_dimensions
    assert clipped.offset == original.offset


def test_clip_inplace():
    raw = np.array([1, 2, 3])
    original = test_module.VoxelData(raw, voxel_dimensions=(2,), offset=(10,))
    original.clip(bbox=((11,), (15,)), na_value=-1, inplace=True)
    assert_array_equal(original.raw, [-1, 2, -1])


def test_clip_empty():
    raw = np.array([1, 2, 3])
    original = test_module.VoxelData(raw, voxel_dimensions=(2,), offset=(10,))
    assert_raises(
        VoxcellError,
        original.clip,
        bbox=((10,), (10,))
    )


def test_clip_out_of_bounds():
    raw = np.array([1, 2, 3])
    original = test_module.VoxelData(raw, voxel_dimensions=(2,), offset=(10,))
    clipped = original.clip(bbox=((-1000,), (1000,)))
    assert_array_equal(clipped.raw, original.raw)


def test_filter():
    raw = np.array([[[11], [12]], [[21], [22]]])
    original = test_module.VoxelData(raw, voxel_dimensions=(2, 6, 10), offset=(10, 20, 30))
    filtered = original.filter(lambda p: np.logical_and(p[:, 0] > 12, p[:, 1] > 26))
    assert_array_equal(original.raw, raw)
    assert_array_equal(filtered.raw, [[[0], [0]], [[0], [22]]])
    assert_array_equal(filtered.offset, original.offset)
    assert_array_equal(filtered.voxel_dimensions, original.voxel_dimensions)


def test_filter_inplace():
    original = test_module.VoxelData(np.array([[[11], [12]], [[21], [22]]]), (2, 6, 10), offset=(10, 20, 30))
    original.filter(lambda p: np.logical_and(p[:, 0] > 12, p[:, 1] > 26), inplace=True)
    assert_array_equal(original.raw, [[[0], [0]], [[0], [22]]])


def test_compact():
    raw = np.array([0, 42, -1])
    original = test_module.VoxelData(raw, voxel_dimensions=(2,), offset=(10,))
    compact = original.compact(na_values=(0, -1))
    assert_array_equal(original.raw, raw)
    assert original.offset == 10
    assert compact.raw == raw[1:2]
    assert compact.voxel_dimensions == original.voxel_dimensions
    assert compact.offset == 12


def test_compact_inplace():
    raw = np.array([0, 42, -1])
    original = test_module.VoxelData(raw, voxel_dimensions=(2,), offset=(10,))
    original.compact(na_values=(0, -1), inplace=True)
    assert original.raw == raw[1:2]
    assert original.offset == 12


def test_orientation_field():
    field = test_module.OrientationField(np.array([[1., 0., 0., 0.]]), voxel_dimensions=(2,))
    npt.assert_almost_equal(
        field.lookup([1.]),
        [np.identity(3)]
    )


def test_orientation_field_compact():
    field = test_module.OrientationField(np.array([[127, 0, 0, 0]], dtype=np.int8), voxel_dimensions=(2,))
    npt.assert_almost_equal(
        field.lookup([1.]),
        [np.identity(3)]
    )


def test_orientation_field_raises():
    with pytest.raises(VoxcellError):
        test_module.OrientationField(np.zeros(4), voxel_dimensions=(1,))
    with pytest.raises(VoxcellError):
        test_module.OrientationField(np.zeros((3, 3)), voxel_dimensions=(1,))


def test_roi_mask():
    field = test_module.ROIMask(np.array([1, 0, 0, 0], dtype=np.uint8), voxel_dimensions=(2,))
    actual = field.lookup([[1.], [3]])
    assert actual.dtype == 'bool'
    assert_array_equal(actual, [True, False])


def test_roi_mask_raises():
    with pytest.raises(VoxcellError, match=re.escape("Invalid dtype: 'int64' (expected: '(u)int8")):
        test_module.ROIMask(np.zeros(4, dtype=np.int64), voxel_dimensions=(1,))


def test_values_to_region_attribute():
    region_map = Mock()
    region_map.get.side_effect = lambda _id, attr: {0: "CA1", 1: "SO", 2: "SP"}[_id]
    values = np.array([1, 0, 0, 1])
    actual = test_module.values_to_region_attribute(values, region_map)
    assert region_map.get.call_count == 2  # called once for each different looked up id
    assert region_map.get.call_args_list == [call(0, attr='acronym'), call(1, attr='acronym')]
    assert np.issubdtype(actual.dtype, str)
    assert_array_equal(actual, ["SO", "CA1", "CA1", "SO"])


def test_values_to_hemisphere():
    values = np.array([2, 1, 1, 2])
    actual = test_module.values_to_hemisphere(values)
    assert np.issubdtype(actual.dtype, str)
    assert_array_equal(actual, ["right", "left", "left", "right"])


def test_values_to_hemisphere_raises_invalid_value():
    values = np.array([2, 1, 1, 99])
    with pytest.raises(VoxcellError, match=re.escape("Invalid values, only [0, 1, 2] are allowed")):
        test_module.values_to_hemisphere(values)


def test_reduce():
    # Add 3 int arrays
    a = test_module.VoxelData(np.array([[11, 12], [21, 22]]), (2, 3))
    b = test_module.VoxelData(np.array([[1, 0], [0, 0]]), (2, 3))
    c = test_module.VoxelData(np.array([[1, 0], [0, 0]]), (2, 3))

    # 1 elements: function is not applied
    d = test_module.VoxelData.reduce(lambda x: x * 100, [a])
    assert_almost_equal(d.raw, [[11, 12], [21, 22]])

    # 2 elements
    e = test_module.VoxelData.reduce(operator.add, [a, b])
    assert_almost_equal(e.raw, [[12, 12], [21, 22]])

    # 3 elements
    f = test_module.VoxelData.reduce(operator.add, [a, b, c])
    assert_almost_equal(f.raw, [[13, 12], [21, 22]])


def test_offset_and_voxel_dimensions_type():
    voxel_data = test_module.VoxelData(
        np.ones((2, 2, 2)), offset=(1, 2, 3), voxel_dimensions=(1, 1, 1))
    assert voxel_data.offset.dtype == np.float32
    assert voxel_data.voxel_dimensions.dtype == np.float32

    voxel_data = test_module.VoxelData(np.ones((2, 2, 2)), voxel_dimensions=(3, 4, 5))
    assert voxel_data.offset.dtype == np.float32


def test_ValueToIndexVoxels():
    values = np.array([[1., 1., 1.], [1., 2., 2.], [3., 3., 3.]])
    br = np.array([[1, 1, 1], [1, 2, 2], [3, 3, 3]], order='C')
    vtiv = test_module.ValueToIndexVoxels(br)

    npt.assert_array_equal(vtiv.value_to_1d_indices(1), [0, 1, 2, 3])
    npt.assert_array_equal(vtiv.value_to_1d_indices(2), [4, 5])
    npt.assert_array_equal(vtiv.value_to_1d_indices(3), [6, 7, 8])
    npt.assert_array_equal(vtiv.value_to_1d_indices(4), [])

    npt.assert_array_equal(vtiv.value_to_indices(1), [[0, 0], [0, 1], [0, 2], [1, 0]])
    npt.assert_array_equal(vtiv.value_to_indices(2), [[1, 1], [1, 2]])
    npt.assert_array_equal(vtiv.value_to_indices(3), [[2, 0], [2, 1], [2, 2]])
    npt.assert_array_equal(
        vtiv.value_to_indices(4),
        np.zeros_like([], shape=(0, 2), dtype=vtiv._indices.dtype),
    )
    npt.assert_array_equal(
        vtiv.value_to_indices(range(1, 5)),
        [[0, 0], [0, 1], [0, 2], [1, 0], [1, 1], [1, 2], [2, 0], [2, 1], [2, 2]],
    )
    npt.assert_array_equal(
        vtiv.value_to_indices(range(10, 15)),  # All values are missing in the input array
        np.zeros_like([], shape=(0, 2), dtype=vtiv._indices.dtype),
    )

    assert vtiv.index_size == 3
    assert vtiv.index_dtype == np.int64
    assert list(vtiv.values) == [1, 2, 3]

    for order in ('K', 'A', 'C', 'F'):
        r = vtiv.ravel(np.array(values, order=order))
        npt.assert_array_equal(r[vtiv.value_to_1d_indices(1)], [1., 1., 1., 1.])

    br = np.array([[1, 1, 1], [1, 2, 2], [3, 3, 3]], order='F')
    vtiv = test_module.ValueToIndexVoxels(br)

    npt.assert_array_equal(vtiv.value_to_1d_indices(1), [0, 1, 3, 6])
    npt.assert_array_equal(vtiv.value_to_1d_indices(2), [4, 7])
    npt.assert_array_equal(vtiv.value_to_1d_indices(3), [2, 5, 8])
    npt.assert_array_equal(vtiv.value_to_1d_indices(4), [])

    npt.assert_array_equal(vtiv.value_to_indices(1), [[0, 0], [1, 0], [0, 1], [0, 2]])
    npt.assert_array_equal(vtiv.value_to_indices(2), [[1, 1], [1, 2]])
    npt.assert_array_equal(vtiv.value_to_indices(3), [[2, 0], [2, 1], [2, 2]])
    npt.assert_array_equal(
        vtiv.value_to_indices(4),
        np.zeros_like([], shape=(0, 2), dtype=vtiv._indices.dtype),
    )
    npt.assert_array_equal(
        vtiv.value_to_indices(range(1, 5)),
        [[0, 0], [1, 0], [0, 1], [0, 2], [1, 1], [1, 2], [2, 0], [2, 1], [2, 2]],
    )
    npt.assert_array_equal(
        vtiv.value_to_indices(range(10, 15)),  # All values are missing in the input array
        np.zeros_like([], shape=(0, 2), dtype=vtiv._indices.dtype),
    )

    for order in ('K', 'A', 'C', 'F'):
        r = vtiv.ravel(np.array(values, order=order))
        npt.assert_array_equal(r[vtiv.value_to_1d_indices(1)], [1., 1., 1., 1.])


def test_ValueToIndexVoxels__raveling():

    # c-contiguous index values
    index_values = np.array([[1., 2., 3.], [4., 5., 6.], [7., 8., 9.]])
    index = test_module.ValueToIndexVoxels(index_values)
    assert index._order == "C"
    
    # c-contiguous volume to ravel with a c-order index
    c_values = np.array([[1., 2., 3.], [4., 5., 6.], [7., 8., 9.]])
    c_raveled = index.ravel(c_values)
    c_unraveled = index.unravel(c_raveled)

    # check the reconstruction is correct
    npt.assert_array_equal(c_unraveled, c_values)

    # should not create a copy
    assert c_raveled.base is c_values
    assert c_unraveled.base is c_values

    # f-contiguous volume to ravel with a c-order index
    f_values = np.asfortranarray(c_values)
    f_raveled = index.ravel(f_values)
    f_unraveled = index.unravel(f_raveled)

    # check the reconstruction is correct
    npt.assert_array_equal(f_unraveled, f_values)

    # should create a copy
    assert f_raveled.base is None
    # should not create a copy because 1d array are both C and F contiguous
    assert f_unraveled.base is f_raveled

    # f-contiguous index values
    index_values = np.asfortranarray(index_values)
    index = test_module.ValueToIndexVoxels(index_values)
    assert index._order == "F"

    # c-contiguous volume to ravel with a f-order index
    c_values = np.array([[1., 2., 3.], [4., 5., 6.], [7., 8., 9.]])
    c_raveled = index.ravel(c_values)
    c_unraveled = index.unravel(c_raveled)

    # check the reconstruction is correct
    npt.assert_array_equal(c_unraveled, c_values)

    # should create a copy
    assert c_raveled.base is None
    # should not create a copy because 1d array are both C and F contiguous
    assert c_unraveled.base is c_raveled

    # f-contiguous volume to ravel with a f-order index
    f_values = np.asfortranarray(c_values)
    f_raveled = index.ravel(f_values)
    f_unraveled = index.unravel(f_raveled)

    # check the reconstruction is correct
    npt.assert_array_equal(f_unraveled, f_values)

    # should not create a copy
    assert f_raveled.base is f_values
    assert f_unraveled.base is f_values


def test_ValueToIndexVoxels__raises():

    index_values = np.array([[1., 2., 3.], [4., 5., 6.], [7., 8., 9.]])
    index = test_module.ValueToIndexVoxels(index_values)

    data = np.array([[0., 1.], [1., 2]])

    with pytest.raises(VoxcellError, match="Shape mismatch"):
        index.ravel(data)

    data = np.array([0., 1.])

    with pytest.raises(VoxcellError, match="Array size mismatch"):
        index.unravel(data)



"""Cell collection access / writer."""
import collections

import h5py
import numpy as np
import pandas as pd

from voxcell.exceptions import VoxcellError
from voxcell.math_utils import angles_to_matrices, euler2mat, mat2euler
from voxcell.quaternion import matrices_to_quaternions, quaternions_to_matrices


def _load_sonata_orientations(group, cells):
    """Set CellCollection rotation matrices from a sonata file."""
    dataset_names = [name for name in list(group) if isinstance(group[name], h5py.Dataset)]
    if not dataset_names:
        return None
    size = group[dataset_names[0]].shape[0]

    def _get_values(properties):
        """Retrieve prop from the group if exists."""
        return [group.get(prop, np.zeros((size,))) for prop in properties]

    props = np.array([
        'orientation_x',
        'orientation_y',
        'orientation_z',
        'orientation_w',
    ])
    orientation_count = np.count_nonzero(np.isin(props, list(group)))
    if orientation_count == 4:
        cells.orientation_format = "quaternions"
        cells.orientations = quaternions_to_matrices(np.vstack(_get_values(props)).T)
    elif orientation_count in [1, 2, 3]:
        raise VoxcellError(
            "Missing orientation fields. Should be 4 quaternions or some euler angles or nothing")
    else:
        # need to keep this rotation_angle ordering for euler2mat (expects z, y, x)
        props = np.array([
            'rotation_angle_zaxis',
            'rotation_angle_yaxis',
            'rotation_angle_xaxis',
        ])
        cells.orientation_format = "eulers"
        cells.orientations = euler2mat(*_get_values(props))


def _load_property(properties, name, values, library_group=None):
    """Loads single property with respect to a library group if presented.

    Args:
        properties (pd.DataFrame): properties
        name (str): property name
        values (array-like): property values
        library_group (h5py.Group): library group
    """
    if values.dtype == object:
        values = values.asstr()
    values = values[()]
    if library_group is not None and name in library_group:
        if library_group[name].dtype == object:
            unique_values = library_group[name].asstr()[()]
        else:
            unique_values = library_group[name][()]

        if unique_values.size < 0.5 * values.size:
            properties[name] = pd.Categorical.from_codes(values, categories=unique_values)
        else:
            properties[name] = unique_values[values]
    else:
        properties[name] = values


def _is_string_enum(series):
    """Whether ``series`` contains enum of strings."""
    is_cat_str = (isinstance(series.dtype, pd.CategoricalDtype) and
                  series.dtype.categories.dtype == object)
    return series.dtype == object or is_cat_str


class CellCollection:
    """Encapsulates all the data related to a collection of cells that compose a circuit.

    Multi-dimensional properties (such as positions and orientations) are attributes.
    General properties are a in a pandas DataFrame object "properties".
    """

    # properties that start with it are dynamic, and handled appropriately, see `dynamics_params` in
    # https://github.com/AllenInstitute/sonata/blob/master/docs/SONATA_DEVELOPER_GUIDE.md#representing-nodes
    SONATA_DYNAMIC_PROPERTY = '@dynamics:'

    def __init__(self, population_name='default', orientation_format="quaternions"):
        """Init CellCollection.

        Args:
            population_name: SONATA population name, currently assume a single population collection
            orientation_format: quaternions or eulers.
        """
        self.population_name = population_name
        self.positions = None
        self.orientations = None
        self.properties = pd.DataFrame()
        self._orientation_format = orientation_format

    def _nonzero_sizes(self):
        return list({len(obj) for obj in [self.properties, self.positions, self.orientations] if
                     obj is not None and len(obj) != 0})

    def _check_sizes(self):
        if len(self._nonzero_sizes()) > 1:
            raise VoxcellError("Lengths of properties, position, orientation don't match")

    def size(self):
        """Return the size of the CellCollection.

        If the positions, orientations or properties are filled with values it checks the sizes of
        these objects and then return the size of the CellCollection.
        """
        self._check_sizes()
        sizes = self._nonzero_sizes()
        return sizes[0] if sizes else 0

    def __len__(self):
        """Return the length of the CellCollection."""
        return self.size()

    @property
    def orientation_format(self):
        """Return the format of the orientation either "eulers" or "quaternions"."""
        return self._orientation_format

    @orientation_format.setter
    def orientation_format(self, val):
        """Set the format of orientation with only "eulers" or "quaternions"."""
        if val not in ["quaternions", "eulers"]:
            raise VoxcellError('You must set orientation_type to either "quaternions" or "eulers".')
        self._orientation_format = val

    def add_properties(self, new_properties, overwrite=True):
        """Adds new columns to the properties DataFrame.

        Args:
            new_properties: a pandas DataFrame object
            overwrite: if True, overwrites columns with the same name.
            Otherwise, a VoxcellError is raised.
        """
        for name, prop in new_properties.items():
            if (not overwrite) and (name in self.properties):
                raise VoxcellError(f"Column '{name}' already exists")
            self.properties[name] = prop

    def remove_unassigned_cells(self):
        """Remove cells with one or more unassigned property."""
        idx_unassigned = self.properties[self.properties.isnull().any(axis=1)].index
        self.properties = self.properties.drop(idx_unassigned)
        self.properties.reset_index(inplace=True, drop=True)
        if self.orientations is not None:
            self.orientations = np.delete(self.orientations, idx_unassigned, 0)
        if self.positions is not None:
            self.positions = np.delete(self.positions, idx_unassigned, 0)

    def as_dataframe(self, index_offset=1):
        """Return a dataframe with all cell properties.

        Args:
            index_offset: index offset (0 or 1). The default may change to 0 in a future version.
        """
        result = self.properties.copy()
        if self.positions is not None:
            result['x'] = self.positions[:, 0]
            result['y'] = self.positions[:, 1]
            result['z'] = self.positions[:, 2]
        if self.orientations is not None:
            result['orientation'] = list(self.orientations)

        result.index = np.arange(index_offset, len(result) + index_offset)

        result.columns = map(str, result.columns)

        return result

    @classmethod
    def from_dataframe(cls, df, index_offset=1):
        """Return a CellCollection object from a dataframe of cell properties.

        Args:
            df: Pandas DataFrame containing the cell properties, with index starting from 0 or 1.
            index_offset: index offset (0 or 1). The default may change to 0 in a future version.
        """
        if not (df.index == np.arange(index_offset, len(df) + index_offset)).all():
            raise VoxcellError(
                f"Index != {index_offset}..{len(df) + index_offset - 1} (got: {df.index.values})"
            )
        result = cls()
        if 'x' in df:
            result.positions = df[['x', 'y', 'z']].values
        if 'orientation' in df:
            result.orientations = np.stack(df['orientation'])
        # don't use `set` for filtering because it looses the order of columns => the restored Cells
        # will have different columns order which is bad.
        props = [column for column in df.columns if column not in ['x', 'y', 'z', 'orientation']]
        result.properties = df[props].reset_index(drop=True)
        return result

    def save(self, filename):
        """Saves this cell collection to HDF5 file in MVD3 or SONATA format.

        Args:
            filename: filepath to write. If it ends with '.mvd3' then it is treated as MVD3,
                otherwise as SONATA.
        """
        none_properties = self.properties.isnull().any(axis=0)
        if none_properties.any():
            names = none_properties.index[none_properties].to_list()
            raise VoxcellError(f"Replace `None` in {names} properties before saving")
        if str(filename).lower().endswith('mvd3'):
            self.save_mvd3(filename)
        else:
            self.save_sonata(filename)

    def save_mvd3(self, filename):
        """Save this cell collection to mvd3 HDF5.

        Args:
            filename(str): fullpath to filename to write
        """
        self._check_sizes()
        with h5py.File(filename, 'w') as f:
            f.create_group('cells')
            f.create_group('library')

            if self.positions is not None:
                f.create_dataset('cells/positions', data=self.positions)

            if self.orientations is not None:
                f.create_dataset('cells/orientations',
                                 data=matrices_to_quaternions(self.orientations))
            # numpy's `np.object` type must be represented as `str_dt`
            # http://docs.h5py.org/en/latest/strings.html
            str_dt = h5py.special_dtype(vlen=str)
            for name, series in self.properties.items():
                values = series.to_numpy()
                if _is_string_enum(series) and not name.startswith(self.SONATA_DYNAMIC_PROPERTY):
                    unique_values, indices = np.unique(values, return_inverse=True)
                    f.create_dataset('cells/properties/' + name, data=indices.astype(np.uint32))
                    f.create_dataset('library/' + name, data=unique_values, dtype=str_dt)
                else:
                    dt = str_dt if values.dtype == object else values.dtype
                    f.create_dataset('cells/properties/' + name, data=values, dtype=dt)

    @classmethod
    def load(cls, filename):
        """Loads CellCollection from a file.

        Args:
            filename: filepath to cells file. If it ends with '.mvd3' then it is treated as
                MVD3 circuit, otherwise as SONATA.

        Returns:
            CellCollection: loaded cells
        """
        if str(filename).lower().endswith('mvd3'):
            return cls.load_mvd3(filename)
        return cls.load_sonata(filename)

    @classmethod
    def load_mvd2(cls, filename):
        """Load a cell collection from mvd2 HDF5.

        This method is a copy of `loadMVD2` from bluepy/v2/impl/cells_mvd.py

        Args:
            filename(str): fullpath to filename to read

        Returns:
            CellCollection object
        """

        def parse_neuron_line(line):
            """Parser for neurons."""
            tokens = line.split()
            return {
                'morphology': tokens[0],
                'hypercolumn': int(tokens[2]),
                'minicolumn': int(tokens[3]),
                'layer': 1 + int(tokens[4]),
                'mtype': mtypes[int(tokens[5])],
                'morph_class': morph_classes[int(tokens[5])],
                'synapse_class': synapse_classes[int(tokens[5])],
                'etype': etypes[int(tokens[6])],
                'x': float(tokens[7]),
                'y': float(tokens[8]),
                'z': float(tokens[9]),
                'orientation': float(tokens[10]),
                'me_combo': tokens[11]
            }

        SECTIONS = (
            "Neurons Loaded", "MicroBox Data",
            "MiniColumnsPosition", "CircuitSeeds",
            "MorphTypes", "ElectroTypes", "FOOTER"
        )

        CATEGORICAL_PROPS = [
            'mtype', 'etype', 'morph_class', 'synapse_class',
            'morphology', 'me_combo'
        ]

        section_lines = collections.defaultdict(list)
        accumulator = None

        with open(filename, 'r', encoding='utf-8') as f:
            for line in f:
                line = line.strip()
                if line in SECTIONS:
                    accumulator = section_lines[line]
                elif accumulator is not None:
                    accumulator.append(line)

        mtypes = [line.split()[0] for line in section_lines['MorphTypes']]
        morph_classes = [line.split()[1] for line in section_lines['MorphTypes']]
        synapse_classes = [line.split()[2] for line in section_lines['MorphTypes']]
        etypes = [line.split()[0] for line in section_lines['ElectroTypes']]

        result = pd.DataFrame(list(map(parse_neuron_line, section_lines['Neurons Loaded'])))
        for c in CATEGORICAL_PROPS:
            result[c] = result[c].astype('category')

        result['orientation'] = list(angles_to_matrices(np.pi * result['orientation'] / 180, 'y'))

        result.index = 1 + np.arange(len(result))
        return cls.from_dataframe(result)

    @classmethod
    def load_mvd3(cls, filename):
        """Load a cell collection from mvd3 HDF5.

        Args:
            filename(str): fullpath to filename to read

        Returns:
            CellCollection object
        """
        cells = cls()

        with h5py.File(filename, 'r') as f:
            data = f['cells']
            if 'positions' in data:
                cells.positions = np.array(data['positions'])

            if 'orientations' in data:
                cells.orientations = np.array(data['orientations'])
                cells.orientations = quaternions_to_matrices(cells.orientations)

            if 'properties' in data:
                for name, values in data['properties'].items():
                    _load_property(cells.properties, name, values, f.get('library'))
        return cells

    def save_sonata(self, filename, forced_library=None, mode='w'):
        """Save this cell collection to sonata HDF5.

        Args:
            filename(str): fullpath to filename to write
            forced_library(iterable of str): names of properties that are
            forced to become part of the @library
            mode(str): mode used to create/open file; passed directly to h5py.File:
                'w' overwrites, 'a' appends

        Note:
          * Only properties that contain strings can be included in the @library
          * when forced_library is None, properties that are categorical are
            included in the @library, unless their number of unique values is
            more than half of all the values
        """
        # pylint: disable=too-many-locals
        forced_library = set() if forced_library is None else set(forced_library)

        self._check_sizes()
        with h5py.File(filename, mode) as h5f:
            population = h5f.create_group(f'/nodes/{self.population_name}')
            population.create_dataset('node_type_id', data=np.full(len(self.properties), -1))
            group = population.create_group('0')
            str_dt = h5py.special_dtype(vlen=str)
            for name, series in self.properties.items():
                if name.startswith(self.SONATA_DYNAMIC_PROPERTY):
                    name = name.split(self.SONATA_DYNAMIC_PROPERTY)[1]
                    dt = str_dt if series.dtype == object else series.dtype
                    group.create_dataset(
                        f'dynamics_params/{name}',
                        data=series.to_numpy(),
                        dtype=dt,
                    )
                elif _is_string_enum(series) or (series.dtype == object and name in forced_library):
                    indices, unique_values = series.factorize()
                    if name in forced_library or len(unique_values) < .5 * len(indices):
                        group.create_dataset(name, data=indices.astype(np.uint32))
                        group.create_dataset(f'@library/{name}', data=unique_values, dtype=str_dt)
                    else:
                        group.create_dataset(name, data=series.to_numpy(), dtype=str_dt)
                else:
                    group.create_dataset(name, data=series.to_numpy())

            if self.orientations is not None:
                if self.orientation_format == "quaternions":
                    quaternions = matrices_to_quaternions(self.orientations)
                    group.create_dataset('orientation_x', data=quaternions[:, 0])
                    group.create_dataset('orientation_y', data=quaternions[:, 1])
                    group.create_dataset('orientation_z', data=quaternions[:, 2])
                    group.create_dataset('orientation_w', data=quaternions[:, 3])
                elif self.orientation_format == "eulers":
                    az, ay, ax = mat2euler(self.orientations)
                    group.create_dataset('rotation_angle_xaxis', data=ax)
                    group.create_dataset('rotation_angle_yaxis', data=ay)
                    group.create_dataset('rotation_angle_zaxis', data=az)

            if self.positions is not None:
                group.create_dataset('x', data=self.positions[:, 0])
                group.create_dataset('y', data=self.positions[:, 1])
                group.create_dataset('z', data=self.positions[:, 2])

    @classmethod
    def load_sonata(cls, filename, population_name=None):
        """Loads a cell collection from sonata HDF5.

        Args:
            filename(str): fullpath to filename to read

        Returns:
            CellCollection object
        """
        cells = cls()

        with h5py.File(filename, 'r') as h5f:
            if population_name is None:
                population_names = list(h5f['/nodes'].keys())
                assert len(population_names) == 1, 'Single population is supported only'
                population_name = population_names[0]

            population = h5f[f'/nodes/{population_name}']
            cells.population_name = population_name

            assert '0' in population, 'Single group "0" is supported only'
            group = population['0']

            keys = set(group.keys())
            if 'x' in group:
                cells.positions = np.vstack((group['x'], group['y'], group['z'])).T

            rotation_datasets = {'orientation_x', 'orientation_y', 'orientation_z', 'orientation_w',
                                 'rotation_angle_zaxis', 'rotation_angle_yaxis',
                                 'rotation_angle_xaxis'}
            if len(rotation_datasets - keys) != len(rotation_datasets):
                _load_sonata_orientations(group, cells)
            properties_names = keys - {'x', 'y', 'z'}.union(rotation_datasets)
            for name in properties_names:
                if not isinstance(group[name], h5py.Dataset):
                    continue
                _load_property(cells.properties, name, group[name], group.get('@library'))

            if 'dynamics_params' in group:
                for name, values in group['dynamics_params'].items():
                    if not isinstance(values, h5py.Dataset):
                        continue
                    if values.dtype == object:
                        values = values.asstr()
                    cells.properties[cls.SONATA_DYNAMIC_PROPERTY + name] = values[()]

        return cells

    def __str__(self):
        """Return the string describing the CellCollection."""
        properties = list(self.properties.columns)

        if self.positions is not None:
            properties += list('xyz')

        if self.orientations is not None:
            if self.orientation_format == "quaternions":
                properties += ['orientation_x',
                               'orientation_y',
                               'orientation_z',
                               'orientation_w',
                               ]
            elif self.orientation_format == "eulers":
                properties += ['rotation_angle_xaxis',
                               'rotation_angle_yaxis',
                               'rotation_angle_zaxis',
                               ]

        return (f'CellCollection[population_name: {self.population_name}]: '
                f'properties: {properties} '
                f'orientation_format: {self.orientation_format} '
                f'count: {len(self.properties)}'
                )

    __repr__ = __str__



"""Quaternions to matrices and vice versa."""

import numpy as np

from voxcell.math_utils import normalize


def matrices_to_quaternions(matrices):
    """Build quaternions from an array of 3x3 rotation matrices.

    Based on multibranch algorithm described here:
    http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/index.htm

    Args:
        matrices: A Nx3x3 numpy array containing N rotation matrices.

    Returns:
        A Nx4 numpy array containing a unit quaternion for each rotation matrix.
        The quaternion components are stored as (x, y, z, w)
    """
    # this is the same algorithm used internally by THREE.js
    quat = np.zeros(shape=matrices.shape[:-2] + (4,), dtype=np.float64)

    # using boolean array "mask" to simulate the branching
    # the boolean array "done" allows us to do the equivalent of "else"
    done = np.zeros(shape=matrices.shape[:-2], dtype=bool)

    # branch 1

    trace = np.trace(matrices, axis1=-2, axis2=-1)
    mask = trace > 0
    done |= mask
    m = matrices[mask]

    s = 0.5 / np.sqrt(trace[mask] + 1.0)
    x = (m[..., 2, 1] - m[..., 1, 2]) * s
    y = (m[..., 0, 2] - m[..., 2, 0]) * s
    z = (m[..., 1, 0] - m[..., 0, 1]) * s
    w = 0.25 / s
    quat[mask] = np.array([x, y, z, w]).transpose()

    # branch 2

    mask = (~done &
            (matrices[..., 0, 0] > matrices[..., 1, 1]) &
            (matrices[..., 0, 0] > matrices[..., 2, 2]))
    done |= mask
    m = matrices[mask]

    s = 2.0 * np.sqrt(1.0 + m[..., 0, 0] - m[..., 1, 1] - m[..., 2, 2])
    x = 0.25 * s
    y = (m[..., 0, 1] + m[..., 1, 0]) / s
    z = (m[..., 0, 2] + m[..., 2, 0]) / s
    w = (m[..., 2, 1] - m[..., 1, 2]) / s
    quat[mask] = np.array([x, y, z, w]).transpose()

    # branch 3

    mask = ~done & (matrices[..., 1, 1] > matrices[..., 2, 2])
    done |= mask
    m = matrices[mask]

    s = 2.0 * np.sqrt(1.0 + m[..., 1, 1] - m[..., 0, 0] - m[..., 2, 2])
    x = (m[..., 0, 1] + m[..., 1, 0]) / s
    y = 0.25 * s
    z = (m[..., 1, 2] + m[..., 2, 1]) / s
    w = (m[..., 0, 2] - m[..., 2, 0]) / s
    quat[mask] = np.array([x, y, z, w]).transpose()

    # branch 4

    mask = ~done
    m = matrices[mask]

    s = 2.0 * np.sqrt(1.0 + m[..., 2, 2] - m[..., 0, 0] - m[..., 1, 1])
    x = (m[..., 0, 2] + m[..., 2, 0]) / s
    y = (m[..., 1, 2] + m[..., 2, 1]) / s
    z = 0.25 * s
    w = (m[..., 1, 0] - m[..., 0, 1]) / s
    quat[mask] = np.array([x, y, z, w]).transpose()

    return quat / np.sqrt((quat ** 2).sum(-1))[..., np.newaxis]


def quaternions_to_matrices(q):
    """Build 3x3 rotation matrices from an array of quaternions.

    Based on algorigthm described here:
    http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToMatrix/index.htm

    Args:
        q: A Nx4 numpy array containing a quaternion for each rotation matrix.
        The quaternion components are stored as (x, y, z, w)

    Returns:
        A Nx3x3 numpy array containing N rotation matrices.
    """
    q = normalize(q.astype(float))

    m11 = 1 - 2 * np.square(q[..., 1]) - 2 * np.square(q[..., 2])
    m12 = 2 * q[..., 0] * q[..., 1] - 2 * q[..., 2] * q[..., 3]
    m13 = 2 * q[..., 0] * q[..., 2] + 2 * q[..., 1] * q[..., 3]

    m21 = 2 * q[..., 0] * q[..., 1] + 2 * q[..., 2] * q[..., 3]
    m22 = 1 - 2 * np.square(q[..., 0]) - 2 * np.square(q[..., 2])
    m23 = 2 * q[..., 1] * q[..., 2] - 2 * q[..., 0] * q[..., 3]

    m31 = 2 * q[..., 0] * q[..., 2] - 2 * q[..., 1] * q[..., 3]
    m32 = 2 * q[..., 1] * q[..., 2] + 2 * q[..., 0] * q[..., 3]
    m33 = 1 - 2 * np.square(q[..., 0]) - 2 * np.square(q[..., 1])

    m = np.vstack([m11, m12, m13, m21, m22, m23, m31, m32, m33]).transpose()

    return m.reshape(m.shape[:-1] + (3, 3))



"""Deprecation utilities."""

import warnings

from voxcell.exceptions import VoxcellError


class VoxcellDeprecationWarning(UserWarning):
    """Voxcell deprecation warning."""


class VoxcellDeprecationError(VoxcellError):
    """Voxcell deprecation error."""


def fail(msg=None):
    """Raise a deprecation exception."""
    raise VoxcellDeprecationError(msg)


def warn(msg=None):
    """Issue a deprecation warning."""
    warnings.warn(msg, VoxcellDeprecationWarning)



"""Access to VoxelBrain.

https://nip.humanbrainproject.eu/documentation/user-manual.html#voxel-brain
"""

import abc
import json
import os
import urllib

import numpy as np
import requests

from voxcell import RegionMap, VoxelData, math_utils
from voxcell.exceptions import VoxcellError


def _download_file(url, filepath, overwrite, allow_empty=False):
    """Download file from `url` if it is missing."""
    if os.path.exists(filepath) and not overwrite:
        return filepath

    tmp_filepath = filepath + ".download"
    try:
        resp = requests.get(url, timeout=None)
        resp.raise_for_status()
        if not (allow_empty or resp.content):
            raise VoxcellError("Empty content")
        with open(tmp_filepath, "wb") as f:
            f.write(resp.content)
        os.rename(tmp_filepath, filepath)
    finally:
        try:
            os.unlink(tmp_filepath)
        except OSError:
            pass

    return filepath


class Atlas:
    """Helper class for atlas access."""
    __metaclass__ = abc.ABCMeta

    def __init__(self):
        """Init Atlas."""
        self._memcache = {}

    @staticmethod
    def open(url, cache_dir=None):
        """Get Atlas object to access atlas stored at URL."""
        parsed = urllib.parse.urlsplit(url)
        if parsed.scheme in ('', 'file'):
            return LocalAtlas(url)

        if parsed.scheme in ('http', 'https'):
            if not parsed.path.startswith('/api/analytics/atlas/releases/'):
                raise VoxcellError(f"Unexpected URL: '{url}'")

            if cache_dir is None:
                raise VoxcellError("`cache_dir` should be specified")

            return VoxelBrainAtlas(url, cache_dir)

        raise VoxcellError(f"Unexpected URL: '{url}'")

    @abc.abstractmethod
    def fetch_data(self, data_type):
        """Fetch `data_type` NRRD."""

    @abc.abstractmethod
    def fetch_hierarchy(self):
        """Fetch brain region hierarchy JSON."""

    def _check_cache(self, key, callback, memcache):
        if key in self._memcache:
            return self._memcache[key]
        result = callback()
        if memcache:
            self._memcache[key] = result
        return result

    def load_data(self, data_type, cls=VoxelData, memcache=False):
        """Load atlas data layer."""
        def _callback():
            return cls.load_nrrd(self.fetch_data(data_type))

        return self._check_cache(
            ('data', data_type, cls),
            callback=_callback,
            memcache=memcache
        )

    def load_region_map(self, memcache=False):
        """Load brain region hierarchy as RegionMap."""
        def _callback():
            return RegionMap.load_json(self.fetch_hierarchy())

        return self._check_cache(
            ('region_map',),
            callback=_callback,
            memcache=memcache
        )

    def get_region_mask(self, value, attr='acronym', with_descendants=True,
                        ignore_case=False, memcache=False):
        """Get VoxelData with 0/1 mask indicating regions matching `value`."""

        def _callback():
            rmap = self.load_region_map()
            brain_regions = self.load_data('brain_regions')
            region_ids = rmap.find(
                value, attr=attr, with_descendants=with_descendants,
                ignore_case=ignore_case
            )
            if not region_ids:
                raise VoxcellError(f"Region not found: '{value}'")
            result = math_utils.isin(brain_regions.raw, region_ids)
            return brain_regions.with_data(result)

        return self._check_cache(
            ('region_mask', value, attr, with_descendants),
            callback=_callback,
            memcache=memcache
        )


class VoxelBrainAtlas(Atlas):
    """Helper class for VoxelBrain atlas."""
    def __init__(self, url, cache_dir):
        """Init VoxelBrainAtlas."""
        super().__init__()
        self._url = url.rstrip("/")
        resp = requests.get(self._url, timeout=None)
        resp.raise_for_status()
        atlas_id = resp.json()[0]['id']
        assert self._url.endswith(atlas_id)
        self._cache_dir = os.path.join(cache_dir, atlas_id)
        if not os.path.exists(self._cache_dir):
            os.makedirs(self._cache_dir)

    def fetch_data(self, data_type):
        """Fetch `data_type` NRRD."""
        resp = requests.get(self._url + "/data", timeout=None)
        resp.raise_for_status()
        data_types = []
        for item in resp.json():
            if item['data_type'] == data_type:
                url = item['url']
                break
            data_types.append(item['data_type'])
        else:
            raise VoxcellError(
                # pylint: disable=consider-using-f-string
                "`data_type` should be one of ({0}), provided: {1}".format(
                    ",".join(data_types), data_type
                )
            )
        filepath = os.path.join(self._cache_dir, f"{data_type}.nrrd")
        return _download_file(url, filepath, overwrite=False)

    def fetch_hierarchy(self):
        """Fetch brain region hierarchy JSON."""
        url = self._url + "/filters/brain_region/65535"
        filepath = os.path.join(self._cache_dir, "hierarchy.json")
        return _download_file(url, filepath, overwrite=False)


class LocalAtlas(Atlas):
    """Helper class for locally stored atlas."""
    def __init__(self, dirpath):
        """Init LocalAtlas."""
        super().__init__()
        self.dirpath = dirpath

    def _get_filepath(self, filename):
        result = os.path.join(self.dirpath, filename)
        if not os.path.exists(result):
            raise VoxcellError(f"File not found: '{result}'")
        return result

    def fetch_data(self, data_type):
        """Return filepath to `data_type` NRRD."""
        return self._get_filepath(f"{data_type}.nrrd")

    def fetch_hierarchy(self):
        """Return filepath to brain region hierarchy JSON."""
        return self._get_filepath("hierarchy.json")

    def fetch_metadata(self):
        """Return filepath to metadata JSON."""
        return self._get_filepath("metadata.json")

    def load_metadata(self, memcache=False):
        """Load brain region metadata as dict."""

        def _callback():
            with open(self.fetch_metadata(), 'r', encoding='utf-8') as f:
                return json.load(f)

        return self._check_cache(('metadata',), callback=_callback, memcache=memcache)

    def get_layers(self):
        """Retrieve and cache the identifiers of each layer of a laminar brain region.

        For each layer of an annotated brain volume, the function returns the set of identifiers of
        all regions included in this layer.

        Note: this function relies on the existence of the files
            * hierarchy.json
            * metadata.json

        See get_layer for the description of the metadata.json file.

        Returns: tuple (names, ids) where `names` is a list of layer names where `ids` is a list
            of sets and. Each set contains the identifiers (ints) of the corresponding layer name
            (str).

        Raises: VoxcellError
            * if the hierarchy file or the metadata file doesn't exist.
            * if metadata.json doesn't contain the key "layers"
            * if the value of "layers" doesn't contain all the required keys:
                "names", "queries" and "attribute".
            * if the value  of "names" or "ids" is not a list, or if these objects are two lists
                of different lengths.
        """

        def _callback():
            metadata = self.load_metadata()

            if 'layers' not in metadata:
                raise VoxcellError('Missing "layers" key')
            layers = metadata['layers']

            if not all(
                key in metadata['layers'] for key in ['names', 'queries', 'attribute']
            ):
                err_msg = (
                    'Missing some "layers" key. The "layers" dictionary has '
                    'the following mandatory keys: "names", "queries" and "attribute"'
                )
                raise VoxcellError(err_msg)

            if not (
                isinstance(layers['names'], list)
                and isinstance(layers['queries'], list)
                and len(layers['names']) == len(layers['queries'])
            ):
                raise VoxcellError(
                    'The values of "names" and "queries" must be lists of the same length'
                )

            region_map = self.load_region_map()

            ids = [
                region_map.find(query, attr=layers['attribute'], with_descendants=True)
                for query in layers['queries']
            ]

            return (layers['names'], ids)

        return self._check_cache(
            ('get_layer_ids'),
            callback=_callback,
            memcache=True,
        )

    def get_layer(self, layer_index):
        """Retrieve the identifiers of a specified layer in laminar brain region.

        Given a layer of an annotated brain volume, the function returns the set of identifiers of
        all regions included in this layer.

        Note: this function relies on the existence of the files
            * hierarchy.json
            * metadata.json

        The content of metadata.json must be of the following form::

            {
                ...
                "layers": {
                    "names": [
                        "layer 1", "layer 2/3", "layer 4", "layer 5", "layer 6", "Olfactory areas"
                    ],
                    "queries": ["@.*1$", "@.*2/3$", "@.*4$", "@.*5$", ".*6[a|b]?$", "OLF"],
                    "attribute": "acronym"
                },
                ...
            }

        The strings of the `queries` list are in one-to-one correspondence with layer `names`.
        The layer `names` list is a user-defined string list which can vary depending on which
        hierarchy.json is used and which brain region is under scrutiny.
        Each query string is used to retrieve the region identifiers of the corresponding
        layer by means of the Atlas RegionMap object instantiated from the hierarchy.json file.
        The syntax of a query string is the same as the one in use for RegionMap.find:
        if the string starts with the symbol '@', the remainder is interpreted as a regular
        expression. Otherwise it is a plain string value and RegionMap.find looks for a
        character-wise full match.
        The value of `attribute` is `acronym` or `name`. This unique value applies for every query
        string.

        Returns: tuple (name, ids) where `name` is the name of the layer (str) with index
            `layer_index` and `ids` is the set of region identifiers of every region
            in this layer according to hierarchy.json.

        Raises: VoxcellError
            * if the hierarchy file or the metadata file doesn't exist.
            * if metadata.json doesn't contain the key "layers"
            * if the value of "layers" doesn't contain all the required keys:
                "names", "queries" and "attribute".
            * if the value  of "names" or "ids" is not a list, or if these objects are two lists
                of different lengths.
        """
        layer_names, layer_ids = self.get_layers()

        return (layer_names[layer_index], layer_ids[layer_index])

    def get_layer_volume(self, memcache=False):
        """Get VoxelData whose voxels are labeled with the layer indices of `brain_regions` voxels.

        Layer indices range from 1 to `number of layers` - 1. The layers of the atlas are defined
        within metadata.json as a list of RegionMap.find query strings.

        If two layer definitions in metadata.json involve the same region identifier,
        the voxels bearing this identifier will be labled with the largest layer index.

        Args:
            memcache: If True, use cache, Otherwise re-compute the layer volume.
                Defaults to False.

        Returns:
            VoxelData object with the same shape and same metadata as brain_regions. Voxels have
            uint8 labels (char type) and each voxel label corresponds to the voxel layer index.
        """

        def _callback():
            brain_regions = self.load_data('brain_regions')
            _, layer_ids = self.get_layers()
            layers = np.zeros_like(brain_regions.raw, dtype=np.uint8)
            for index, ids in enumerate(layer_ids, 1):
                mask = math_utils.isin(brain_regions.raw, ids)
                layers[mask] = index

            return brain_regions.with_data(layers)

        return self._check_cache(
            ('layer_volume'),
            callback=_callback,
            memcache=memcache,
        )



"""Voxcell."""

from voxcell.cell_collection import CellCollection
from voxcell.exceptions import VoxcellError
from voxcell.region_map import RegionMap
from voxcell.voxel_data import (
    OrientationField,
    ROIMask,
    VoxelData,
    values_to_hemisphere,
    values_to_region_attribute,
)



from collections import namedtuple
from docutils.parsers.rst import Directive, directives
from docutils import nodes


class bbp_table_value(nodes.Element):
    pass


class BBPTableValue(Directive):
    has_content = True
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec = {'type': directives.unchanged,
                   'value': directives.unchanged,
                   'comment': directives.unchanged
                   }

    def run(self):
        resultnode = bbp_table_value()
        self.options['field'] = self.arguments[0]
        resultnode.options = self.options
        return [resultnode]


BBPTableSectionInfo = namedtuple('BBPTableSectionInfo',
                                   'name, description, docname, targetid')


class bbp_table_section(nodes.General, nodes.Element):
    pass


class BBPTableSection(Directive):
    has_content = True
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec = {'description': directives.unchanged,
                   }

    def _make_table(self, rows):
        def create_table_row(row_cells):
            row = nodes.row()
            for cell in row_cells:
                row += nodes.entry('', nodes.paragraph(text=cell))
            return row

        header = ('Field', 'Value Type', 'Suggested Value', 'Comments' )
        colwidths = (1, 1, 2, 3)

        assert len(header) == len(colwidths)
        tgroup = nodes.tgroup(cols=len(header))
        for c in colwidths:
            tgroup += nodes.colspec(colwidth=c)
        tgroup += nodes.thead('', create_table_row(header))
        tbody = nodes.tbody()
        tgroup += tbody
        for row in rows:
            tbody += create_table_row((row.options['field'],
                                       row.options['type'],
                                       row.options['value'],
                                       row.options['comment']))

        table = nodes.table('', tgroup)
        return table

    def run(self):
        env = self.state.document.settings.env
        if not hasattr(env, 'all_bbp_table_sections'):
            env.all_bbp_table_sections = []

        name = self.arguments[0]
        description = self.options['description']
        targetid = "bbp_tablesection-%d" % env.new_serialno('bbp_table_section')

        node = nodes.Element()
        self.state.nested_parse(self.content, self.content_offset, node)

        section_info = BBPTableSectionInfo(name, description, env.docname, targetid)
        env.all_bbp_table_sections.append(section_info)

        children = []
        for child in node:
            if isinstance(child, bbp_table_value):
                children.append(child)

        resultnode = nodes.section(ids=[targetid])
        resultnode += [nodes.title(text=name),
                       nodes.paragraph(text=description),
                       self._make_table(children),
                       nodes.line()
                       ]

        return [resultnode]


class bbp_table_section_index(nodes.Element):
    pass


class BBPTableSectionIndex(Directive):
    '''create a place-holder for an index'''

    def run(self):
        return [bbp_table_section_index('')]


def process_bbp_table_section_index(app, doctree, fromdocname):
    env = app.builder.env

    for node in doctree.traverse(bbp_table_section_index):
        references = []
        for section in env.all_bbp_table_sections:
            ref = nodes.reference(section.name, section.name)
            ref['refdocname'] = section.docname
            ref['refuri'] = app.builder.get_relative_uri(
                fromdocname, section.docname) + '#' + section.targetid

            para = nodes.paragraph('', '', ref)
            item = nodes.list_item('', para, nodes.paragraph(text=section.description))
            references.append(item)

        content = nodes.bullet_list('', *references)
        node.replace_self(content)


def setup(app):
    app.add_node(bbp_table_section_index)
    app.add_directive('bbp_table_section_index', BBPTableSectionIndex)

    app.add_node(bbp_table_section)
    app.add_directive('bbp_table_section', BBPTableSection)
    app.add_config_value('bbp_table_section', {}, 'env')

    app.add_node(bbp_table_value)
    app.add_directive('bbp_table_value', BBPTableValue)

    app.connect('doctree-resolved', process_bbp_table_section_index)



import numpy as np
import numpy.testing as npt
import pytest

import voxcell.math_utils as test_module
from voxcell import VoxelData


def test_clip():
    r = test_module.clip(
        np.array([
            [0, 0, 0],
            [0, 1, 0],
            [0, 0, 0]
        ]),
        (np.array([1, 1]), np.array([1, 1]))
    )
    npt.assert_equal(r, np.array([[1]]))


def test_is_diagonal_true():
    A = np.array([
        [2, 0],
        [0, 3]
    ])
    assert test_module.is_diagonal(A)


def test_is_diagonal_false():
    A = np.array([
        [2, 0],
        [1, 3]
    ])
    assert not test_module.is_diagonal(A)


def test_lcmm():
    npt.assert_equal(12, test_module.lcmm([2, 3, 4]))


def test_angles_to_matrices_x():
    angles = [np.pi / 2]
    expected = [[
        [1, 0, 0],
        [0, 0, -1],
        [0, 1, 0],
    ]]
    result = test_module.angles_to_matrices(angles, 'x')
    npt.assert_almost_equal(expected, result)


def test_angles_to_matrices_y():
    angles = [np.pi / 2]
    expected = [[
        [0, 0, 1],
        [0, 1, 0],
        [-1, 0, 0],
    ]]
    result = test_module.angles_to_matrices(angles, 'y')
    npt.assert_almost_equal(expected, result)


def test_angles_to_matrices_z():
    angles = [np.pi / 2]
    expected = [[
        [0, -1, 0],
        [1, 0, 0],
        [0, 0, 1],
    ]]
    result = test_module.angles_to_matrices(angles, 'z')
    npt.assert_almost_equal(expected, result)


def test_normalize_empty():
    npt.assert_equal(test_module.normalize([]), [])


def test_normalize_0():
    npt.assert_equal(test_module.normalize([1, 0, 0]), [1, 0, 0])


def test_normalize_1():
    npt.assert_equal(test_module.normalize([2, 2, 1]), [2./3, 2./3, 1./3])


def test_normalize_3():
    npt.assert_equal(test_module.normalize([[1, 0, 0], [0, 0, 0]]), [[1, 0, 0], [0, 0, 0]])


def test_isin_empty():
    npt.assert_equal(test_module.isin([], [1, 2]), [])


def test_isin_0():
    npt.assert_equal(
        test_module.isin(
            [[1, 2], [2, 3]],
            []
        ),
        [[False, False], [False, False]]
    )


def test_isin_1():
    npt.assert_equal(
        test_module.isin(
            [[1, 2], [2, 3]],
            [2, 4, 4]
        ),
        [[False, True], [True, False]]
    )


def test_isin_2():
    npt.assert_equal(
        test_module.isin(
            [[1, 2], [2, 3]],
            set([1, 2, 3, 5])
        ),
        [[True, True], [True, True]]
    )


def test_euler2mat():
    pi2 = np.pi / 2
    pi3 = np.pi / 3
    pi4 = np.pi / 4
    pi6 = np.pi / 6
    actual = test_module.euler2mat(
        [0.0, pi2, pi3],  # rotation_angle_z
        [pi2, 0.0, pi4],  # rotation_angle_y
        [pi2, pi2, pi6],  # rotation_angle_x
    )
    expected = np.array([
        [
            [0., 0., 1.],
            [1., 0., 0.],
            [0., 1., 0.],
        ],
        [
            [0., -1.,  0.],
            [0.,  0., -1.],
            [1.,  0.,  0.],
        ],
        [
            [0.35355339, -0.61237244,  0.70710678],
            [0.92677670,  0.12682648, -0.35355339],
            [0.12682648,  0.78033009,  0.61237244],
        ]
    ])
    npt.assert_almost_equal(actual, expected)


def test_mat2euler_roundtrip():
    original = np.asarray([
        [
            # ay = pi / 2
            [0., 0., 1.],
            [1., 0., 0.],
            [0., 1., 0.],
        ],
        [
            # ay = -pi / 2
            [ 0., 0., -1.],
            [ 0., 1.,  0.],
            [ 1., 0.,  0.],
        ],
        [
            # ay = pi / 4
            [0.35355339, -0.61237244,  0.70710678],
            [0.92677670,  0.12682648, -0.35355339],
            [0.12682648,  0.78033009,  0.61237244],
        ]
    ])
    actual = test_module.euler2mat(
        *test_module.mat2euler(original)
    )
    npt.assert_almost_equal(original, actual)


def _random_angles(n):
    return 2 * np.pi * np.random.random(n)


def test_mat2euler_roundtrip_random():
    n = 100
    az = _random_angles(n)
    ay = _random_angles(n)
    ax = _random_angles(n)
    mm = test_module.euler2mat(az, ay, ax)
    actual = test_module.euler2mat(
        *test_module.mat2euler(mm)
    )


def test_mat2euler_raises_1():
    with pytest.raises(AssertionError):
        test_module.mat2euler(np.asarray([
            [0., -1.,  0.],
            [0.,  0., -1.],
            [1.,  0.,  0.],
        ]))


def test_mat2euler_raises_2():
    with pytest.raises(AssertionError):
        test_module.mat2euler(np.asarray([
            [
                [0., -1.],
                [1.,  0.],
            ]
        ]))


@pytest.fixture
def simple_voxel_data():
    return VoxelData(np.ones((3, 4, 5)),
                     voxel_dimensions=[1, 1, 1],
                     offset=[0, 0, 0])


@pytest.mark.parametrize(
    "with_sub_segments",
    [
        pytest.param(True, id="Get indices and cut the segment according to voxel planes"),
        pytest.param(False, id="Get indices only"),
    ],
)
@pytest.mark.parametrize(
    "segment, expected_indices, expected_sub_segments",
    [
        pytest.param(
            [[-1, -1, -1], [-2, -2, -2]],
            np.zeros((0, 3)),
            [[-1, -1, -1, -2, -2, -2]],
            id="The segment does not intersect any voxel",
        ),
        pytest.param(
            [[1.5, 0, 3], [0, 2.5, 0]],
            [
                [1, 0, 3],
                [1, 0, 2],
                [0, 0, 1],
                [0, 1, 1],
                [0, 1, 0],
                [0, 2, 0],
            ],
            [
                [1.5, 0.0, 3.0, 1.5, 0.0, 3.0],
                [1.5, 0.0, 3.0, 1.0, 0.83333333, 2.0],
                [1.0, 0.83333333, 2.0, 0.9, 1.0, 1.8],
                [0.9, 1.0, 1.8, 0.5, 1.66666667, 1.0],
                [0.5, 1.66666667, 1.0, 0.3, 2.0, 0.6],
                [0.3, 2.0, 0.6, 0.0, 2.5, 0.0],
            ],
            id="The segment intersects several voxels",
        ),
        pytest.param(
            [[0.25, 0.25, 0.25], [0.75, 0.75, 0.75]],
            [[0, 0, 0]],
            [
                [0.25, 0.25, 0.25, 0.75, 0.75, 0.75],
            ],
            id="The segment is entirely inside the voxel",
        ),
        pytest.param(
            [[0, 0.5, 0.5], [0.5, 0.5, 0.5]],
            [[0, 0, 0]],
            [
                [0.0, 0.5, 0.5, 0.5, 0.5, 0.5],
            ],
            id="The segment touches the xmin plane and is inside the first voxel",
        ),
        pytest.param(
            [[1, 0.25, 0.25], [1, 0.75, 0.75]],
            [[1, 0, 0]],
            [
                [1, 0.25, 0.25, 1, 0.75, 0.75],
            ],
            id="The segment is contained in the xmin plane",
        ),
        pytest.param(
            [[1, 0.5, 0.5], [1.5, 0.5, 0.5]],
            [[1, 0, 0]],
            [
                [1.0, 0.5, 0.5, 1.5, 0.5, 0.5],
            ],
            id="The segment touches the xmin plane and is inside the second voxel",
        ),
        pytest.param(
            [[0.5, 0.5, 0.5], [1, 0.5, 0.5]],
            [[0, 0, 0], [1, 0, 0]],
            [
                [0.5, 0.5, 0.5, 1, 0.5, 0.5],
                [1, 0.5, 0.5, 1, 0.5, 0.5],
            ],
            id="The segment touches the xmax plane and is inside the voxel",
        ),
        pytest.param(
            [[0.5, 0, 0.5], [0.5, 0.5, 0.5]],
            [[0, 0, 0]],
            [
                [0.5, 0.0, 0.5, 0.5, 0.5, 0.5],
            ],
            id="The segment touches the ymin plane and is inside the first voxel",
        ),
        pytest.param(
            [[0.5, 1, 0.5], [0.5, 1.5, 0.5]],
            [[0, 1, 0]],
            [
                [0.5, 1.0, 0.5, 0.5, 1.5, 0.5],
            ],
            id="The segment touches the ymin plane and is inside the second voxel",
        ),
        pytest.param(
            [[0.5, 0.5, 0.5], [0.5, 1, 0.5]],
            [[0, 0, 0], [0, 1, 0]],
            [
                [0.5, 0.5, 0.5, 0.5, 1.0, 0.5],
                [0.5, 1.0, 0.5, 0.5, 1.0, 0.5],
            ],
            id="The segment touches the ymax plane and is inside the voxel",
        ),
        pytest.param(
            [[0.25, 1, 0.25], [0.75, 1, 0.75]],
            [[0, 1, 0]],
            [
                [0.25, 1.0, 0.25, 0.75, 1.0, 0.75],
            ],
            id="The segment is contained in the ymin plane",
        ),
        pytest.param(
            [[0.5, 0.5, 0], [0.5, 0.5, 0.5]],
            [[0, 0, 0]],
            [
                [0.5, 0.5, 0.0, 0.5, 0.5, 0.5],
            ],
            id="The segment touches the zmin plane and is inside the first voxel",
        ),
        pytest.param(
            [[0.5, 0.5, 1], [0.5, 0.5, 1.5]],
            [[0, 0, 1]],
            [
                [0.5, 0.5, 1.0, 0.5, 0.5, 1.5],
            ],
            id="The segment touches the zmin plane and is inside the second voxel",
        ),
        pytest.param(
            [[0.5, 0.5, 0.5], [0.5, 0.5, 1]],
            [[0, 0, 0], [0, 0, 1]],
            [
                [0.5, 0.5, 0.5, 0.5, 0.5, 1.0],
                [0.5, 0.5, 1.0, 0.5, 0.5, 1.0],
            ],
            id="The segment touches the zmax plane and is inside the voxel",
        ),
        pytest.param(
            [[0.25, 0.25, 1], [0.75, 0.75, 1]],
            [[0, 0, 1]],
            [
                [0.25, 0.25, 1.0, 0.75, 0.75, 1.0],
            ],
            id="The segment is contained in the zmin plane",
        ),
        pytest.param(
            [[0.25, 0.25, 0.25], [1, 1, 1]],
            [[0, 0, 0], [1, 1, 1]],
            [
                [0.25, 0.25, 0.25, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
            ],
            id="The segment touches the corner of several voxels and is inside one voxel",
        ),
        pytest.param(
            [[0.25, 0.25, 0.25], [2.25, 0.25, 0.25]],
            [[0, 0, 0], [1, 0, 0], [2, 0, 0]],
            [
                [0.25, 0.25, 0.25, 1.0, 0.25, 0.25],
                [1.0, 0.25, 0.25, 2.0, 0.25, 0.25],
                [2.0, 0.25, 0.25, 2.25, 0.25, 0.25],
            ],
            id="The segment crosses several voxels along X",
        ),
        pytest.param(
            [[0.25, 0.25, 0.25], [0.25, 2.25, 0.25]],
            [[0, 0, 0], [0, 1, 0], [0, 2, 0]],
            [
                [0.25, 0.25, 0.25, 0.25, 1.0, 0.25],
                [0.25, 1.0, 0.25, 0.25, 2.0, 0.25],
                [0.25, 2.0, 0.25, 0.25, 2.25, 0.25],
            ],
            id="The segment crosses several voxels along Y",
        ),
        pytest.param(
            [[0.25, 0.25, 0.25], [0.25, 0.25, 2.25]],
            [[0, 0, 0], [0, 0, 1], [0, 0, 2]],
            [
                [0.25, 0.25, 0.25, 0.25, 0.25, 1.0],
                [0.25, 0.25, 1.0, 0.25, 0.25, 2.0],
                [0.25, 0.25, 2.0, 0.25, 0.25, 2.25],
            ],
            id="The segment crosses several voxels along Z",
        ),
        pytest.param(
            [[0, 1, 1], [3, 1, 1]],
            [
                [0, 1, 1],
                [1, 1, 1],
                [2, 1, 1],
            ],
            [
                [0.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 2.0, 1.0, 1.0],
                [2.0, 1.0, 1.0, 3.0, 1.0, 1.0],
            ],
            id="The segment touches the boundaries of several voxels along X",
        ),
        pytest.param(
            [[3, 1, 1], [0, 1, 1]],
            [
                [2, 1, 1],
                [1, 1, 1],
                [0, 1, 1],
            ],
            [
                [3.0, 1.0, 1.0, 2.0, 1.0, 1.0],
                [2.0, 1.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 0.0, 1.0, 1.0],
            ],
            id="The segment touches the boundaries of several voxels along X and is reversed",
        ),
        pytest.param(
            [[1, 0, 1], [1, 3, 1]],
            [
                [1, 0, 1],
                [1, 1, 1],
                [1, 2, 1],
                [1, 3, 1],
            ],
            [
                [1.0, 0.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0, 2.0, 1.0],
                [1.0, 2.0, 1.0, 1.0, 3.0, 1.0],
                [1.0, 3.0, 1.0, 1.0, 3.0, 1.0],
            ],
            id="The segment touches the boundaries of several voxels along Y",
        ),
        pytest.param(
            [[1, 3, 1], [1, 0, 1]],
            [
                [1, 3, 1],
                [1, 2, 1],
                [1, 1, 1],
                [1, 0, 1],
            ],
            [
                [1.0, 3.0, 1.0, 1.0, 3.0, 1.0],
                [1.0, 3.0, 1.0, 1.0, 2.0, 1.0],
                [1.0, 2.0, 1.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0, 0.0, 1.0],
            ],
            id="The segment touches the boundaries of several voxels along Y and is reversed",
        ),
        pytest.param(
            [[1, 1, 0], [1, 1, 3]],
            [
                [1, 1, 0],
                [1, 1, 1],
                [1, 1, 2],
                [1, 1, 3],
            ],
            [
                [1.0, 1.0, 0.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0, 1.0, 2.0],
                [1.0, 1.0, 2.0, 1.0, 1.0, 3.0],
                [1.0, 1.0, 3.0, 1.0, 1.0, 3.0],
            ],
            id="The segment touches the boundaries of several voxels along Z",
        ),
        pytest.param(
            [[1, 1, 3], [1, 1, 0]],
            [
                [1, 1, 3],
                [1, 1, 2],
                [1, 1, 1],
                [1, 1, 0],
            ],
            [
                [1.0, 1.0, 3.0, 1.0, 1.0, 3.0],
                [1.0, 1.0, 3.0, 1.0, 1.0, 2.0],
                [1.0, 1.0, 2.0, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 1.0, 1.0, 0.0],
            ],
            id="The segment touches the boundaries of several voxels along Z and is reversed",
        ),
        pytest.param(
            [[1.5, 1.5, 1.5], [0.5, 0.5, 0.5]],
            [
                [1, 1, 1],
                [0, 0, 0],
            ],
            [
                [1.5, 1.5, 1.5, 1.0, 1.0, 1.0],
                [1.0, 1.0, 1.0, 0.5, 0.5, 0.5],
            ],
            id="The segment is oblique and passes only in the upper right corner of the voxel",
        ),
    ],
)
@pytest.mark.parametrize(
    "translation",
    [
        pytest.param(
            None,
            id="No translation",
        ),
        pytest.param(
            [-0.5, -0.5, -0.5],
            id="Slight negative translation",
        ),
        pytest.param(
            [0.5, 0.5, 0.5],
            id="Slight positive translation",
        ),
        pytest.param(
            [-1.5e6, -1.5e6, -1.5e6],
            id="Large negative translation",
        ),
        pytest.param(
            [1.5e6, 1.5e6, 1.5e6],
            id="Large positive translation",
        ),
    ]
)
@pytest.mark.parametrize(
    "scale",
    [
        pytest.param(
            None,
            id="No rescale",
        ),
        pytest.param(
            0.75,
            id="Slight downscale",
        ),
        pytest.param(
            1.5,
            id="Slight upscale",
        ),
        pytest.param(
            1.5e6,
            id="Large upscale",
        ),
    ]
)
def test_voxel_intersection(
    simple_voxel_data,
    segment,
    expected_indices,
    expected_sub_segments,
    with_sub_segments,
    translation,
    scale,
):
    if translation is not None:
        simple_voxel_data.offset += np.array(translation, dtype=simple_voxel_data.offset.dtype)
        segment = (np.array(segment, dtype=float) + np.array(translation, dtype=float)).tolist()
        expected_sub_segments = (
            np.array(expected_sub_segments, dtype=float)
            + np.concatenate([translation] * 2, dtype=float)
        ).tolist()

    if scale is not None:
        simple_voxel_data.voxel_dimensions *= scale
        segment = (
            simple_voxel_data.offset + (
                np.array(segment, dtype=float) - simple_voxel_data.offset
            ) * scale
        ).tolist()
        expected_sub_segments = (
            np.concatenate([simple_voxel_data.offset] * 2) + (
                np.array(expected_sub_segments, dtype=float)
                - np.concatenate([simple_voxel_data.offset] * 2)
            ) * scale
        ).tolist()

    result = test_module.voxel_intersection(
        segment, simple_voxel_data, return_sub_segments=with_sub_segments
    )
    if with_sub_segments:
        indices, sub_segments = result
        np.testing.assert_allclose(sub_segments, expected_sub_segments, atol=abs((scale or 1) / 1e6))
    else:
        indices = result
    np.testing.assert_array_equal(indices, expected_indices)



import pytest

from voxcell.sonata import NodePopulation
from voxcell.utils import deprecate


def test_init():
    with pytest.raises(deprecate.VoxcellDeprecationError):
        NodePopulation(1, 1)



import numpy as np
from numpy.testing import assert_almost_equal, assert_array_equal

import voxcell.quaternion as test_module


def check_normalized(mat):
    assert_almost_equal(np.dot(mat, mat.T)[0][0], 1.0)


def check_matrices_to_quaternions(count, matrix, quaternion):
    quat = test_module.matrices_to_quaternions(np.array([matrix] * count))
    assert_almost_equal(quat, np.array([quaternion] * count), decimal=6)
    check_normalized(quat)


def check_quaternions_to_matrices(count, matrix, quaternion):
    m = test_module.quaternions_to_matrices(np.array([quaternion] * count))
    assert_almost_equal(m, np.array([matrix] * count), decimal=6)


def test_quaternion_empty():
    quat = test_module.matrices_to_quaternions(np.empty((0, 3, 3)))
    assert_array_equal(quat, np.empty((0, 4)))


def test_quaternion_identity():
    m = np.diag([1, 1, 1])
    q = np.array([0, 0, 0, 1])

    check_matrices_to_quaternions(1, m, q)
    check_matrices_to_quaternions(2, m, q)
    check_matrices_to_quaternions(100, m, q)


def test_quaternion_90_y():
    m = np.array([[0, 0, 1],
                  [0, 1, 0],
                  [-1, 0, 0]])

    q = np.array([0., 0.707107, 0., 0.707107])

    check_matrices_to_quaternions(1, m, q)
    check_matrices_to_quaternions(2, m, q)
    check_matrices_to_quaternions(100, m, q)


def test_quaternion_copysign():
    # copysign should not affect our implementaton
    # see:
    # http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/paul.htm
    m = np.diag([1, 1, -1])
    q = np.array([0, 0, 0, 1])

    check_matrices_to_quaternions(1, m, q)
    check_matrices_to_quaternions(2, m, q)
    check_matrices_to_quaternions(100, m, q)


def test_quaternion_180_y():
    m = np.diag([-1, 1, -1])
    q = np.array([0, 1, 0, 0])

    check_matrices_to_quaternions(1, m, q)
    check_matrices_to_quaternions(2, m, q)
    check_matrices_to_quaternions(100, m, q)


def test_quaternion_180_heading_90_attitude():
    m = np.array([[0, 1, 0],
                  [1, 0, 0],
                  [0, 0, -1]])

    q = np.array([0.707107, 0.707107, 0, 0])

    check_matrices_to_quaternions(1, m, q)
    check_matrices_to_quaternions(2, m, q)
    check_matrices_to_quaternions(100, m, q)


def test_matrix_empty():
    quat = test_module.quaternions_to_matrices(np.empty((0, 4)))
    assert_array_equal(quat, np.empty((0, 3, 3)))


def test_matrix_identity():
    m = np.diag([1, 1, 1])
    q = np.array([0, 0, 0, 1])

    check_quaternions_to_matrices(1, m, q)
    check_quaternions_to_matrices(2, m, q)
    check_quaternions_to_matrices(100, m, q)

    # check non-normalized quaternion
    check_quaternions_to_matrices(1, m, 1000 * q)


def test_matrix_90_y():
    m = np.array([[0, 0, 1],
                  [0, 1, 0],
                  [-1, 0, 0]])

    q = np.array([0., 0.707107, 0., 0.707107])

    check_quaternions_to_matrices(1, m, q)
    check_quaternions_to_matrices(2, m, q)
    check_quaternions_to_matrices(100, m, q)

    # check non-normalized quaternion
    check_quaternions_to_matrices(1, m, 1000 * q)


def test_matrix_180_y():
    m = np.diag([-1, 1, -1])
    q = np.array([0, 1, 0, 0])

    check_quaternions_to_matrices(1, m, q)
    check_quaternions_to_matrices(2, m, q)
    check_quaternions_to_matrices(100, m, q)

    # check non-normalized quaternion
    check_quaternions_to_matrices(1, m, 1000 * q)


def test_matrix_180_heading_90_attitude():
    m = np.array([[0, 1, 0],
                  [1, 0, 0],
                  [0, 0, -1]])

    q = np.array([0.707107, 0.707107, 0, 0])

    check_quaternions_to_matrices(1, m, q)
    check_quaternions_to_matrices(2, m, q)
    check_quaternions_to_matrices(100, m, q)

    # check non-normalized quaternion
    check_quaternions_to_matrices(1, m, 1000 * q)


def test_roundtrip_complex():
    # these are all 24 cube rotations taken from:
    # http://www.euclideanspace.com/maths/geometry/rotations/conversions/quaternionToMatrix/examples/index.htm

    series = np.array([
        [[1, 0, 0],
         [0, 1, 0],
         [0, 0, 1]],
        [[0, 0, 1],
         [0, 1, 0],
         [-1, 0, 0]],
        [[-1, 0, 0],
         [0, 1, 0],
         [0, 0, -1]],
        [[0, 0, -1],
         [0, 1, 0],
         [1, 0, 0]],
        [[0, -1, 0],
         [1, 0, 0],
         [0, 0, 1]],
        [[0, 0, 1],
         [1, 0, 0],
         [0, 1, 0]],
        [[0, 1, 0],
         [1, 0, 0],
         [0, 0, -1]],
        [[0, 0, -1],
         [1, 0, 0],
         [0, -1, 0]],
        [[0, 1, 0],
         [-1, 0, 0],
         [0, 0, 1]],
        [[0, 0, 1],
         [-1, 0, 0],
         [0, -1, 0]],
        [[0, -1, 0],
         [-1, 0, 0],
         [0, 0, -1]],
        [[0, 0, -1],
         [-1, 0, 0],
         [0, 1, 0]],
        [[1, 0, 0],
         [0, 0, -1],
         [0, 1, 0]],
        [[0, 1, 0],
         [0, 0, -1],
         [-1, 0, 0]],
        [[-1, 0, 0],
         [0, 0, -1],
         [0, -1, 0]],
        [[0, -1, 0],
         [0, 0, -1],
         [1, 0, 0]],
        [[1, 0, 0],
         [0, -1, 0],
         [0, 0, -1]],
        [[0, 0, -1],
         [0, -1, 0],
         [-1, 0, 0]],
        [[-1, 0, 0],
         [0, -1, 0],
         [0, 0, 1]],
        [[0, 0, 1],
         [0, -1, 0],
         [1, 0, 0]],
        [[1, 0, 0],
         [0, 0, 1],
         [0, -1, 0]],
        [[0, -1, 0],
         [0, 0, 1],
         [-1, 0, 0]],
        [[-1, 0, 0],
         [0, 0, 1],
         [0, 1, 0]],
        [[0, 1, 0],
         [0, 0, 1],
         [1, 0, 0]]],
        dtype=np.float64)

    quat = test_module.matrices_to_quaternions(series)
    restored = test_module.quaternions_to_matrices(quat)
    assert_almost_equal(series, restored, decimal=6)
    check_normalized(quat)



"""Access to volumetric data."""
from functools import reduce

import nrrd
import numpy as np
from numpy.testing import assert_array_equal

from voxcell import math_utils
from voxcell.exceptions import VoxcellError
from voxcell.quaternion import quaternions_to_matrices


def _pivot_axes(a, k):
    """Move `k` first dimensions of `a` to the end, preserving their order.

    I.e., _pivot_axes(A x B x C x D x E, 2) -> C x D x E x A x B
    """
    n = len(a.shape)
    assert 0 <= k <= n
    return np.moveaxis(a, np.arange(k), np.arange(n - k, n))


class VoxelData:
    """Wrap volumetric data and some basic metadata."""

    OUT_OF_BOUNDS = -1

    def __init__(self, raw, voxel_dimensions, offset=None):
        """Note that he units for the metadata will depend on the atlas being used.

        Args:
            raw(numpy.ndarray): actual voxel values
            voxel_dimensions(tuple of numbers): size of each voxel in space.
            offset(tuple of numbers): offset from an external atlas origin
        """
        voxel_dimensions = np.array(voxel_dimensions, dtype=np.float32)
        if len(voxel_dimensions.shape) > 1:
            raise VoxcellError(
                f"voxel_dimensions should be a 1-d array (got: {len(voxel_dimensions.shape)})")
        self.voxel_dimensions = voxel_dimensions

        if offset is None:
            self.offset = np.zeros(self.ndim, dtype=np.float32)
        else:
            offset = np.array(offset, dtype=np.float32)
            if offset.shape != (self.ndim,):
                raise VoxcellError(
                    f"'offset' shape should be: {(self.ndim,)} (got: {offset.shape})")
            self.offset = offset

        if len(raw.shape) < self.ndim:
            raise VoxcellError(
                f"'raw' should have at least {self.ndim} dimensions (got: {len(raw.shape)})")
        self.raw = raw

    @property
    def voxel_volume(self):
        """Voxel volume."""
        return abs(np.prod(self.voxel_dimensions))

    @property
    def ndim(self):
        """Number of dimensions."""
        return len(self.voxel_dimensions)

    @property
    def shape(self):
        """Number of voxels in each dimension."""
        return self.raw.shape[:self.ndim]

    @property
    def payload_shape(self):
        """Shape of the data stored per voxel."""
        return self.raw.shape[self.ndim:]

    @property
    def bbox(self):
        """Bounding box."""
        return np.array([self.offset,
                         self.offset + self.voxel_dimensions * self.shape])

    @classmethod
    def load_nrrd(cls, nrrd_path):
        """Read volumetric data from a nrrd file.

        Args:
            nrrd_path (str|pathlib.Path): path to the nrrd file.
        """
        data, header = nrrd.read(str(nrrd_path))

        # According to http://teem.sourceforge.net/nrrd/format.html#spacedirections,
        # 'space directions' could use 'none' for "payload" axes.
        # As we need space directions only for "space" axes, we rely on either
        # 'space dimension' or 'space' header to slice them out.
        # NB: only a subset of possible 'space' values is supported at the moment.
        if 'space dimension' in header:
            ndim = header['space dimension']
        elif 'space' in header:
            ndim = {
                'right-anterior-superior': 3, 'RAS': 3,
                'left-anterior-superior': 3, 'LAS': 3,
                'left-posterior-superior': 3, 'LPS': 3,
                'posterior-inferior-right': 3, 'PIR': 3,
            }[header['space']]
        else:
            ndim = 0  # use all 'space directions'

        if 'space directions' in header:
            directions = np.array(header['space directions'][-ndim:], dtype=np.float32)
            if not math_utils.is_diagonal(directions):
                raise NotImplementedError("Only diagonal space directions supported at the moment")
            spacings = directions.diagonal()
        elif 'spacings' in header:
            spacings = np.array(header['spacings'][-ndim:], dtype=np.float32)
        else:
            raise VoxcellError("spacings not defined in nrrd")

        offset = None
        if 'space origin' in header:
            offset = np.array(header['space origin'], dtype=np.float32)

        # In NRRD 'payload' axes go first, move them to the end
        raw = _pivot_axes(data, len(data.shape) - len(spacings))

        return cls(raw, spacings, offset)

    def save_nrrd(self, nrrd_path, encoding=None):
        """Save a VoxelData to an nrrd file.

        Args:
            nrrd_path(string|pathlib.Path): full path to nrrd file
            encoding(string): encoding option to save as
        """
        # from http://teem.sourceforge.net/nrrd/format.html#space
        space_directions = np.diag(self.voxel_dimensions)
        dim_defect = len(self.raw.shape) - self.ndim
        if dim_defect > 0:
            # The nrrd specifications require that
            # we prepend a nan array for each of the extra axes
            # before specifying the volume 3D axes.
            # For instance, a volume of orientations (3D direction vectors or quaternions)
            # or of RGB colors (3D int vectors) requires an initial nan array.
            space_directions = [[np.nan] * self.ndim] * dim_defect + list(space_directions)
        header = {
            'space dimension': self.ndim,
            'space directions': space_directions,
            'space origin': self.offset,
        }

        # Case of an N-dimensional vector field over a volume
        if dim_defect == 1 and np.issubdtype(self.raw.dtype, np.number):
            # Required by ITK-SNAP (http://www.itksnap.org/pmwiki/pmwiki.php) and 3D Slicer
            # (https://download.slicer.org/) to render 3D-vector fields using a RGB field.
            # Note that ITK-SNAP raises a "Failed to load segmentation error" but lets you watch
            # the volumetric data.
            header['kinds'] = ['vector', 'domain', 'domain', 'domain']

        if encoding is not None:
            header['encoding'] = encoding

        # In NRRD 'payload' axes should go first, move them to the beginning
        nrrd_data = _pivot_axes(self.raw, self.ndim)
        nrrd.write(str(nrrd_path), nrrd_data, header=header)

    def lookup(self, positions, outer_value=None):
        """Find the values in raw corresponding to the given positions.

        Args:
            positions: list of positions (x, y, z).
            outer_value: value to be returned for positions outside the atlas space.
                If `None`, a VoxcellError is raised in that case.

        Returns:
            Numpy array with the values of the voxels corresponding to each position.
        """
        voxel_idx = self.positions_to_indices(positions, strict=outer_value is None)
        outer_mask = np.any(voxel_idx == VoxelData.OUT_OF_BOUNDS, axis=-1)
        if np.any(outer_mask):
            result = np.full(
                voxel_idx.shape[:-1] + self.payload_shape,
                outer_value,
                dtype=self.raw.dtype
            )
            inner_mask = np.logical_not(outer_mask)  # pylint: disable=assignment-from-no-return
            result[inner_mask] = self._lookup_by_indices(voxel_idx[inner_mask])
        else:
            result = self._lookup_by_indices(voxel_idx)
        return result

    def _lookup_by_indices(self, voxel_idx):
        """Values for the given voxels."""
        voxel_idx_tuple = tuple(voxel_idx.transpose())
        return self.raw[voxel_idx_tuple]

    def positions_to_indices(self, positions, strict=True, keep_fraction=False):
        """Take positions, and the index of the voxel to which they belong.

        Args:
            positions(np.array of Nx3): positions in voxel volume
            strict(bool): raise VoxcellError if any of the positions are out of bounds
            keep_fraction(bool): keep the fractional portion of the positions

        Returns:
            np.array(Nx3) with the voxels coordinates corresponding to each position.

        """
        result = (positions - self.offset) / self.voxel_dimensions
        result[np.abs(result) < 1e-7] = 0.  # suppress rounding errors around 0

        if not keep_fraction:
            result = np.floor(result).astype(int)

        result[result < 0] = VoxelData.OUT_OF_BOUNDS
        result[(result >= self.shape) & (positions >= self.bbox[1])] = VoxelData.OUT_OF_BOUNDS

        if not keep_fraction:
            result = np.clip(result, a_min=None, a_max=np.array(self.shape) - 1)
        else:
            result = np.clip(result, a_min=None, a_max=np.nextafter(self.shape, -1))

        if strict and np.any(result == VoxelData.OUT_OF_BOUNDS):
            raise VoxcellError("Out of bounds position")

        return result

    def indices_to_positions(self, indices):
        """Return positions within given voxels.

        Use fractional indices to obtain positions within voxels
        (for example, index (0.5, 0.5) would give the center of voxel (0, 0)).
        """
        return indices * self.voxel_dimensions + self.offset

    def count(self, values):
        """Number of voxels with value from the given list.

        `values` could be a single value or an iterable.
        """
        if isinstance(values, set):
            # numpy.in1d expects an array-like object as second parameter
            values = list(values)
        return np.count_nonzero(np.in1d(self.raw, values))

    def volume(self, values):
        """Total volume of voxels with value from the given list.

        `values` could be a single value or an iterable.
        """
        return self.count(values) * self.voxel_volume

    def clip(self, bbox, na_value=0, inplace=False):
        """Assign `na_value` to voxels outside of axis-aligned bounding box.

        Args:
            bbox: bounding box in real-world coordinates
            na_value: value to use for voxels outside of bbox
            inplace(bool): modify data inplace

        Returns:
            None if `inplace` is True, new VoxelData otherwise
        """
        bbox = np.array(bbox)
        if bbox.shape != (2, self.ndim):
            raise VoxcellError(f"Invalid bbox shape: {bbox.shape}")

        aabb = ((bbox - self.offset) / self.voxel_dimensions).astype(int)

        # ensure clipped volume is inside bbox
        aa, bb = np.clip(aabb, np.full(self.ndim, -1), self.shape)
        aa += 1
        bb -= 1
        if np.any(aa > bb):
            raise VoxcellError("Empty slice")

        indices = tuple(range(a, b + 1) for a, b in zip(aa, bb))

        if inplace:
            mask = np.full_like(self.raw, False, dtype=bool)
            mask[indices] = True
            self.raw[np.logical_not(mask)] = na_value
            return None

        raw = np.full_like(self.raw, na_value)
        raw[indices] = self.raw[indices]
        return VoxelData(raw, self.voxel_dimensions, self.offset)

    def filter(self, predicate, inplace=False):
        """Set values for voxel positions not satisfying `predicate` to zero.

        Args:
            predicate: N x k [float] -> N x 1 [bool]
            inplace(bool): modify data inplace

        Returns:
            None if `inplace` is True, new VoxelData otherwise
        """
        ijk = np.stack(np.mgrid[[slice(0, d) for d in self.shape]], axis=-1)
        xyz = self.indices_to_positions(0.5 + ijk)
        mask = predicate(xyz.reshape(-1, self.ndim)).reshape(self.shape)

        if inplace:
            self.raw[np.invert(mask)] = 0
            return None

        raw = np.zeros_like(self.raw)
        raw[mask] = self.raw[mask]
        return VoxelData(raw, self.voxel_dimensions, self.offset)

    def compact(self, na_values=(0,), inplace=False):
        """Reduce size of raw data by clipping N/A values.

        Args:
            na_values(tuple): values to clip
            inplace(bool): modify data inplace

        Returns:
            None if `inplace` is True, new VoxelData otherwise
        """
        mask = np.logical_not(  # pylint: disable=assignment-from-no-return
            math_utils.isin(self.raw, na_values)
        )
        aabb = math_utils.minimum_aabb(mask)

        raw = math_utils.clip(self.raw, aabb)
        offset = self.indices_to_positions(aabb[0])

        if inplace:
            self.raw = raw
            self.offset = offset
            return None

        return VoxelData(raw, self.voxel_dimensions, offset)

    def with_data(self, raw):
        """Return VoxelData of the same shape with different data."""
        return VoxelData(raw, self.voxel_dimensions, self.offset)

    @staticmethod
    def reduce(function, iterable):
        """Return a VoxelData by reducing the raw contents of the VoxelData objects in iterable.

        Note: if iterable contains only one item, a copy is returned (but function
        is not applied)

        Args:
            function (Callable[[np.array, np.array], np.array]): the function to be
                applied to numpy arrays
            iterable (Sequence[VoxelData]): a sequence of VoxelData objects
        """
        iterable = list(iterable)
        if not iterable:
            raise TypeError('Attempting to reduce an empty sequence')

        for element in iterable:
            assert isinstance(element, VoxelData)
            assert_array_equal(element.voxel_dimensions, iterable[0].voxel_dimensions)
            assert_array_equal(element.offset, iterable[0].offset)

        return iterable[0].with_data(reduce(function, (x.raw for x in iterable)))


class OrientationField(VoxelData):
    """Volumetric data with rotation per voxel.

    See Also:
        Orientation Field File Format in the documentation
    """
    def __init__(self, *args, **kwargs):
        """Init OrientationField."""
        super().__init__(*args, **kwargs)
        if self.raw.dtype not in (np.int8, np.float32, np.float64):
            raise VoxcellError(f"Invalid volumetric data dtype: {self.raw.dtype}")
        if self.payload_shape != (4,):
            raise VoxcellError("Volumetric data should store (x, y, z, w) tuple per voxel")

    # pylint: disable=arguments-differ
    def lookup(self, positions):
        """Orientations corresponding to the given positions.

        Args:
            positions: list of positions (x, y, z).

        Returns:
            Numpy array with the rotation matrices corresponding to each position.
        """
        result = super().lookup(positions, outer_value=None)

        # normalize int8 data
        if result.dtype == np.int8:
            result = result / 127.0

        # change quaternion component order: (w, x, y, z) -> (x, y, z, w)
        result = np.roll(result, -1, axis=-1)

        return quaternions_to_matrices(result)


class ROIMask(VoxelData):
    """Volumetric data defining 0/1 mask.

    See Also:
        Mask Image for Region of Interest (ROI) in the documentation
    """
    def __init__(self, *args, **kwargs):
        """Init ROIMask."""
        super().__init__(*args, **kwargs)
        if self.raw.dtype not in (np.int8, np.uint8, bool):
            raise VoxcellError(f"Invalid dtype: '{self.raw.dtype}' (expected: '(u)int8')")
        self.raw = self.raw.astype(bool)


class ValueToIndexVoxels:
    """Efficient access to indices of unique values of the values array.

    Useful for when one has an "annotations volume" or "brain region volume" that has
    regions indicated by unique values, and these are used to create masks.  Often,
    it's faster to avoid mask creation, and use indices directly

    Example:
        # To calculate the cell count based on densities of a certain ID in the brain_regions volume
        vtiv = ValueToIndexVoxels(brain_regions.raw)
        density_copy = vtiv.ravel(density.raw.copy())
        indices = vtiv.value_to_1d_indices(value=id_)
        cell_count = np.sum(density_copy[indices]) * voxel_volume)
    """

    def __init__(self, values):
        """Initialize.

        Args:
            values(np.array): volume with each voxel marked with a value; usually to group regions
        """
        self._order = "C" if values.flags["C_CONTIGUOUS"] else "F"
        self._shape = values.shape

        values = values.ravel(order=self._order)
        uniques, counts = np.unique(values, return_counts=True)

        offsets = np.empty(len(counts) + 1, dtype=np.uint64)
        offsets[0] = 0
        offsets[1:] = np.cumsum(counts)

        self._offsets = offsets
        self._indices = np.argsort(values, kind="stable")
        self._mapping = {v: i for i, v in enumerate(uniques)}
        self._index_dtype = values.dtype

    @property
    def index_size(self):
        """Return the size of the unique index values."""
        return len(self._mapping)

    @property
    def index_dtype(self):
        """Return the dytpe of the index values."""
        return self._index_dtype

    @property
    def values(self):
        """Unique values that are found in the original volume."""
        return np.fromiter(self._mapping, dtype=self.index_dtype)

    def value_to_1d_indices(self, value):
        """Return the indices array corresponding to the 'value'.

        Note: These are 1D indices, so the assumption is they are applied to a volume
        who has been ValueToIndexVoxels::ravel(volume)
        """
        if value not in self._mapping:
            return np.array([], dtype=self._indices.dtype)

        group_index = self._mapping[value]
        return self._indices[self._offsets[group_index]:self._offsets[group_index + 1]]

    def value_to_indices(self, values):
        """Return the ND-indices array corresponding to the 'values'.

        This can be convenient to get the positions of the given values in the VoxelData space:
            raw = np.array([[11, 12], [21, 22]])
            v = VoxelData(raw, voxel_dimensions=(2, 3), offset=np.array([2, 2]))
            vtiv = ValueToIndexVoxels(v.raw)
            positions = v.indices_to_positions(vtiv.value_to_indices(11))

        Note: The given 'values' can be given as one scalar value or as a list of values. In both
            case a list of ND-indices will be returned.
        """
        if np.isscalar(values):
            flat_indices = self.value_to_1d_indices(values)
        else:
            flat_indices = np.concatenate(
                [self.value_to_1d_indices(i) for i in values]
            )
        return np.array(
            np.unravel_index(flat_indices, self._shape, order=self._order)
        ).T

    def ravel(self, voxel_data):
        """Ensure `voxel_data` matches the layout that the 1D indices can be used."""
        if voxel_data.shape != self._shape:
            raise VoxcellError(
                f"Shape mismatch:\n"
                f"Index initial shape: {self._shape}\n"
                f"Argument shape: {voxel_data.shape}"
            )
        return voxel_data.ravel(order=self._order)

    def unravel(self, raveled_voxel_array):
        """Ensure `raveled_voxel_array` is reshaped with the contiguous order used to be raveled."""
        if raveled_voxel_array.size != np.prod(self._shape):
            raise VoxcellError(
                "Array size mismatch:\n"
                f"Index initial size: {np.prod(self._shape)}\n"
                f"Argument size: {raveled_voxel_array.size}"
            )
        return raveled_voxel_array.reshape(self._shape, order=self._order)


def values_to_region_attribute(values, region_map, attr="acronym"):
    """Convert region ids to the corresponding region attribute.

    It can be used to convert the values retrieved with `VoxelData.lookup()`.

    Args:
        values (np.array): array containing the values to be converted.
        region_map (RegionMap): instance used to map values to region acronyms.
        attr (str): attribute name to lookup.

    Returns:
        Numpy array with the converted values.

    Raises:
        VoxcellError: if the attribute or any region id is not found.

    See Also:
        Scalar Image File Format in the documentation
    """
    ids, idx = np.unique(values, return_inverse=True)
    resolved = np.array([region_map.get(_id, attr=attr) for _id in ids])
    return resolved[idx]


def values_to_hemisphere(values):
    """Convert integer values 0, 1, 2 to "undefined", "left" and "right" hemisphere labels.

    It can be used to convert the values retrieved with VoxelData.lookup.

    Args:
        values: numpy array containing the values to be converted.

    Returns:
        Numpy array with the converted values.

    Raises:
        VoxcellError: if any of the values is invalid.

    See Also:
        Scalar Image File Format in the documentation
    """
    ids_map = {0: "undefined", 1: "left", 2: "right"}
    ids, idx = np.unique(values, return_inverse=True)
    if not set(ids_map).issuperset(ids):
        raise VoxcellError(f"Invalid values, only {list(ids_map)} are allowed")
    resolved = np.array([ids_map[_id] for _id in ids])
    return resolved[idx]



"""SONATA Nodes access / writer."""
from voxcell.utils import deprecate


# pylint: disable=unused-argument
class NodePopulation:
    """Read / write access to one-population on-group SONATA node collection."""
    def __init__(self, name, size):
        """Init NodePopulation."""
        deprecate.fail("This class has been removed please use CellCollection.load() instead.")
